Package: lobstr
Title: Visualize R Data Structures with Trees
Version: 1.2.2
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: A set of tools for inspecting and understanding R data
    structures inspired by str(). Includes ast() for visualizing abstract
    syntax trees, ref() for showing shared references, cst() for showing
    call stack trees, and obj_size() for computing object sizes.
License: MIT + file LICENSE
URL: https://lobstr.r-lib.org/, https://github.com/r-lib/lobstr
BugReports: https://github.com/r-lib/lobstr/issues
Depends: R (>= 3.6.0)
Imports: crayon, methods, prettyunits, rlang (>= 1.0.0)
Suggests: covr, pillar, pkgdown, testthat (>= 3.0.0)
LinkingTo: cpp11 (>= 0.4.2)
Config/build/compilation-database: true
Config/Needs/website: tidyverse/tidytemplate
Config/roxygen2/version: 8.1.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-08-31 23:00:22 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-09-01 02:20:10 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",lobstr_bytes)
S3method(c,lobstr_bytes)
S3method(format,lobstr_bytes)
S3method(format,lobstr_inspector)
S3method(print,lobstr_bytes)
S3method(print,lobstr_inspector)
S3method(print,lobstr_raw)
S3method(print,lobstr_srcref)
S3method(tree_label,"NULL")
S3method(tree_label,"function")
S3method(tree_label,character)
S3method(tree_label,default)
S3method(tree_label,environment)
S3method(tree_label,lobstr_srcfile_ref)
S3method(tree_label,lobstr_srcref)
S3method(tree_label,lobstr_srcref_location)
S3method(tree_label,srcfile)
S3method(tree_label,srcref)
export(ast)
export(cst)
export(mem_used)
export(obj_addr)
export(obj_addrs)
export(obj_size)
export(obj_sizes)
export(ref)
export(src)
export(sxp)
export(tree)
export(tree_label)
import(rlang)
useDynLib(lobstr, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# lobstr <a href="https://lobstr.r-lib.org"><img src="man/figures/logo.png" align="right" height="138" alt="lobstr website" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/lobstr)](https://cran.r-project.org/package=lobstr)
[![R-CMD-check](https://github.com/r-lib/lobstr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/lobstr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/lobstr/graph/badge.svg)](https://app.codecov.io/gh/r-lib/lobstr)
<!-- badges: end -->

lobstr provides tools in the same vein as `str()`, which allow you to
dig into the detail of an object.

## Installation

Install the released version of lobstr from CRAN:

``` r
install.packages("lobstr")
```

You can install the development version with:

``` r
# install.packages("pak")
pak::pak("r-lib/lobstr")
```

## Example

### Abstract syntax trees

`ast()` draws the abstract syntax tree of R expressions:

``` r
ast(a + b + c)
#> █─`+` 
#> ├─█─`+` 
#> │ ├─a 
#> │ └─b 
#> └─c

ast(function(x = 1) {
  if (x > 0) print("Hi!")
})
#> █─`function` 
#> ├─█─x = 1 
#> ├─█─`{` 
#> │ └─█─`if` 
#> │   ├─█─`>` 
#> │   │ ├─x 
#> │   │ └─0 
#> │   └─█─print 
#> │     └─"Hi!" 
#> └─NULL
```

### References

`ref()` shows hows objects can be shared across data structures by
digging into the underlying \_\_ref\_\_erences:

``` r
x <- 1:1e6
y <- list(x, x, x)
ref(y)
#> █ [1:0x126225d88] <list> 
#> ├─[2:0x1114afb90] <int> 
#> ├─[2:0x1114afb90] 
#> └─[2:0x1114afb90]

e <- rlang::env()
e$self <- e
ref(e)
#> █ [1:0x126563548] <env> 
#> └─self = [1:0x126563548]
```

A related tool is `obj_size()`, which computes the size of an object
taking these shared references into account:

``` r
obj_size(x)
#> 680 B
obj_size(y)
#> 760 B
```

### Call stack trees

`cst()` shows how frames on the call stack are connected:

``` r
f <- function(x) g(x)
g <- function(x) h(x)
h <- function(x) x
f(cst())
#>     ▆
#>  1. ├─f(cst())
#>  2. │ └─g(x)
#>  3. │   └─h(x)
#>  4. └─lobstr::cst()
```

<!-- README.md is generated from README.Rmd. Please edit that file -->

# lobstr <img src="man/figures/logo.png" align="right" height="139" />

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/lobstr)](https://cran.r-project.org/package=lobstr)
[![R-CMD-check](https://github.com/r-lib/lobstr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/lobstr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/lobstr/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/lobstr?branch=main)
<!-- badges: end -->

lobstr provides tools in the same vein as `str()`, which allow you to
dig into the detail of an object.

## Installation

Install the released version of lobstr from CRAN:

``` r
install.packages("lobstr")
```

You can install the development version with:

``` r
# install.packages("devtools")
devtools::install_github("r-lib/lobstr")
```

## Example

### Abstract syntax trees

`ast()` draws the abstract syntax tree of R expressions:

``` r
ast(a + b + c)
#> █─`+` 
#> ├─█─`+` 
#> │ ├─a 
#> │ └─b 
#> └─c

ast(function(x = 1) {
  if (x > 0) print("Hi!")
})
#> █─`function` 
#> ├─█─x = 1 
#> ├─█─`{` 
#> │ └─█─`if` 
#> │   ├─█─`>` 
#> │   │ ├─x 
#> │   │ └─0 
#> │   └─█─print 
#> │     └─"Hi!" 
#> └─<inline srcref>
```

### References

`ref()` shows hows objects can be shared across data structures by
digging into the underlying \_\_ref\_\_erences:

``` r
x <- 1:1e6
y <- list(x, x, x)
ref(y)
#> █ [1:0x7fed114eaea8] <list> 
#> ├─[2:0x7fed21f373b8] <int> 
#> ├─[2:0x7fed21f373b8] 
#> └─[2:0x7fed21f373b8]

e <- rlang::env()
e$self <- e
ref(e)
#> █ [1:0x7fecf1856f00] <env> 
#> └─self = [1:0x7fecf1856f00]
```

A related tool is `obj_size()`, which computes the size of an object
taking these shared references into account:

``` r
obj_size(x)
#> 680 B
obj_size(y)
#> 760 B
```

### Call stack trees

`cst()` shows how frames on the call stack are connected:

``` r
f <- function(x) g(x)
g <- function(x) h(x)
h <- function(x) x
f(cst())
#>     ▆
#>  1. ├─global f(cst())
#>  2. │ └─global g(x)
#>  3. │   └─global h(x)
#>  4. └─lobstr::cst()
```

#' How much memory is currently used by R?
#'
#' `mem_used()` wraps around `gc()` and returns the exact number of bytes
#' currently used by R. Note that changes will not match up exactly to
#' [obj_size()] as session specific state (e.g. [.Last.value]) adds minor
#' variations.
#'
#' @export
#' @examples
#' prev_m <- 0; m <- mem_used(); m - prev_m
#'
#' x <- 1:1e6
#' prev_m <- m; m <- mem_used(); m - prev_m
#' obj_size(x)
#'
#' rm(x)
#' prev_m <- m; m <- mem_used(); m - prev_m
#'
#' prev_m <- m; m <- mem_used(); m - prev_m
mem_used <- function() {
  new_bytes(sum(gc()[, 1] * c(node_size(), 8)))
}

node_size <- function() {
  bit <- 8L * .Machine$sizeof.pointer
  if (!(bit == 32L || bit == 64L)) {
    stop("Unknown architecture", call. = FALSE)
  }

  if (bit == 32L) 28L else 56L
}

#' Pretty tree-like object printing
#'
#' A cleaner and easier to read replacement for `str` for nested list-like
#' objects
#'
#' @param x A tree like object (list, etc.)
#' @param index_unnamed Should children of containers without names have indices
#'   used as stand-in?
#' @param max_depth How far down the tree structure should be printed. E.g. `1`
#'   means only direct children of the root element will be shown. Useful for
#'   very deep lists.
#' @param max_vec_len How many elements should be printed for vectors?
#' @param show_environments Should environments be treated like normal lists and
#'   recursed into?
#' @param hide_scalar_types Should atomic scalars be printed with type and
#'   length like vectors? E.g. `x <- "a"` would be shown as `x<char [1]>: "a"`
#'   instead of `x: "a"`.
#' @param max_length How many elements should be printed? This is useful in case
#'   you try and print an object with 100,000 items in it.
#' @param val_printer Function that values get passed to before being drawn to
#'   screen. Can be used to color or generally style output.
#' @param class_printer Same as `val_printer` but for the the class types of
#'   non-atomic tree elements.
#' @param show_attributes Should attributes be printed as a child of the list or
#'   avoided?
#' @param remove_newlines Should character strings with newlines in them have
#'   the newlines removed? Not doing so will mess up the vertical flow of the
#'   tree but may be desired for some use-cases if newline structure is
#'   important to understanding object state.
#' @param tree_chars List of box characters used to construct tree. Needs
#'   elements `$h` for horizontal bar, `$hd` for dotted horizontal bar, `$v` for
#'   vertical bar, `$vd` for dotted vertical bar, `$l` for l-bend, and `$j` for
#'   junction (or middle child).
#' @param ... Ignored (used to force use of names)
#'
#' @return console output of structure
#'
#' @examples
#'
#' x <- list(
#'   list(id = "a", val = 2),
#'   list(
#'     id = "b",
#'     val = 1,
#'     children = list(
#'       list(id = "b1", val = 2.5),
#'       list(
#'         id = "b2",
#'         val = 8,
#'         children = list(
#'           list(id = "b21", val = 4)
#'         )
#'       )
#'     )
#'   ),
#'   list(
#'     id = "c",
#'     val = 8,
#'     children = list(
#'       list(id = "c1"),
#'       list(id = "c2", val = 1)
#'     )
#'   )
#' )
#'
#' # Basic usage
#' tree(x)
#'
#' # Even cleaner output can be achieved by not printing indices
#' tree(x, index_unnamed = FALSE)
#'
#' # Limit depth if object is potentially very large
#' tree(x, max_depth = 2)
#'
#' # You can customize how the values and classes are printed if desired
#' tree(x, val_printer = function(x) {
#'   paste0("_", x, "_")
#' })
#' @export
tree <- function(
  x,
  ...,
  index_unnamed = FALSE,
  max_depth = 10L,
  max_length = 1000L,
  max_vec_len = 10L,
  show_environments = TRUE,
  hide_scalar_types = TRUE,
  val_printer = crayon::blue,
  class_printer = crayon::silver,
  show_attributes = FALSE,
  remove_newlines = TRUE,
  tree_chars = box_chars()
) {
  rlang::check_dots_empty()

  # Pack up the unchanging arguments into a list and send to tree_internal
  termination_type <- tree_internal(
    x,
    opts = list(
      index_unnamed = index_unnamed,
      max_depth = max_depth,
      max_length = max_length,
      max_vec_len = max_vec_len,
      show_envs = show_environments,
      hide_scalar_types = hide_scalar_types,
      val_printer = val_printer,
      class_printer = class_printer,
      show_attributes = show_attributes,
      remove_newlines = remove_newlines,
      tree_chars = tree_chars
    )
  )
  if (termination_type == "early") {
    cat("...", "\n")
  }

  invisible(x)
}

# Tree printing internal function
#
# This is the internal function for the main tree printing code. It wraps the
# static options arguments from the user-facing `tree()` into a single opts
# list to make recursive calls cleaner. It also has arguments that as it is
# called successively but the end-user shouldn't see or use.
tree_internal <- function(
  x,
  x_id = NULL,
  branch_hist = character(0),
  opts,
  attr_mode = FALSE,
  counter_env = rlang::new_environment(
    data = list(n_printed = 0, envs_seen = c())
  )
) {
  counter_env$n_printed <- counter_env$n_printed + 1
  # Stop if we've reached the max number of times printed desired
  if (counter_env$n_printed > opts$max_length) {
    return("early")
  }
  # Since self-loops can occur in environments check to see if we've seen any
  # environments before
  already_seen <- FALSE
  if (rlang::is_environment(x)) {
    already_seen <- any(vapply(
      counter_env$envs_seen,
      identical,
      x,
      FUN.VALUE = logical(1)
    ))
    if (!already_seen) {
      # If this environment is new, add it to the seen
      counter_env$envs_seen[[length(counter_env$envs_seen) + 1]] <- x
    }
  }

  depth <- length(branch_hist)

  # Build branch string from branch history
  # Start with empty spaces
  branch_chars <- rep_len("  ", depth)

  branch_chars[branch_hist == "child"] <- paste0(opts$tree_chars$v, " ")
  branch_chars[grepl("attr", branch_hist, fixed = TRUE)] <- paste0(
    opts$tree_chars$vd,
    " "
  )

  # Next update the final element (aka the current step) with the correct branch type
  last_step <- branch_hist[depth]
  root_node <- length(branch_hist) == 0

  branch_chars[depth] <- if (root_node) {
    ""
  } else {
    paste0(
      if (grepl("last", last_step)) opts$tree_chars$l else opts$tree_chars$j,
      if (grepl("attribute", last_step)) {
        opts$tree_chars$hd
      } else {
        opts$tree_chars$h
      }
    )
  }
  # Build label
  label <- paste0(
    x_id,
    make_type_abrev(x, opts$hide_scalar_types),
    if (!rlang::is_null(x_id) && x_id != "") ": ",
    tree_label(x, opts),
    if (already_seen) " (Already seen)"
  )

  # Figure out how many children we have (plus attributes if they are being
  # printed) so we can setup how to proceed
  x_attributes <- attributes(x)
  if (attr_mode) {
    # Filter out "names" attribute as this is already shown by tree
    x_attributes <- x_attributes[names(x_attributes) != "names"]
  }
  has_attributes <- length(x_attributes) > 0 && opts$show_attributes

  has_children <- has_attributes || length(x) > 1
  max_depth_reached <- depth >= opts$max_depth && has_children

  # Do the actual printing to the console with an optional ellipses to indicate
  # we've reached the max depth and won't recurse more
  cat(
    paste(branch_chars, collapse = ""),
    label,
    if (max_depth_reached) "...",
    "\n",
    sep = ""
  )

  # ===== Start recursion logic
  if (already_seen || max_depth_reached) {
    return("Normal finish")
  }

  if (rlang::is_list(x) || is_printable_env(x)) {
    # Coerce current object to a plain list. This is necessary as some s3
    # classes override `[[` and return funky stuff like themselves (see s3 class
    # "package_version")
    children <- if (is_printable_env(x)) {
      # Environments are funky as they don't have names before conversion to list
      # but do after, so let them handle their conversion.
      # We use all.names = TRUE in an effort to fully explain the object
      as.list.environment(x, all.names = TRUE)
    } else {
      # By wiping all attributes except for the names we force the object to be
      # a plain list. This is inspired by the (now depreciated) rlang::as_list().
      attributes(x) <- list(names = names(x))
      as.list(x)
    }

    # Traverse children, if any exist
    n_children <- length(children)
    child_names <- names(children)
    # If children have names, give them the names
    for (i in seq_along(children)) {
      id <- child_names[i]
      if ((rlang::is_null(id) || id == "") && opts$index_unnamed) {
        id <- crayon::italic(i)
      }

      child_type <- if (i < n_children) {
        "child"
      } else if (has_attributes) {
        # We use "attrs" here instead of full "attribute" so a grep for
        # attributes just gets plain "attribute" or "last-attribute" but a grep
        # for "attr" gets all attribute related types
        "pre-attrs"
      } else {
        "last-child"
      }
      termination_type <- Recall(
        x = children[[i]],
        x_id = id,
        branch_hist = c(branch_hist, child_type),
        opts = opts,
        counter_env = counter_env
      )
      if (termination_type == "early") {
        return(termination_type)
      }
    }
  }
  # ===== End recursion logic

  # Add any attributes as an "attr" prefixed children at end
  if (has_attributes) {
    n_attributes <- length(x_attributes)
    for (i in seq_len(n_attributes)) {
      termination_type <- Recall(
        x = x_attributes[[i]],
        x_id = crayon::italic(paste0(
          "attr(,\"",
          names(x_attributes)[i],
          "\")"
        )),
        opts = opts,
        branch_hist = c(
          branch_hist,
          paste0(if (i == n_attributes) "last-", "attribute")
        ),
        attr_mode = TRUE, # Let tree know this is an attribute
        counter_env = counter_env
      )
      if (termination_type == "early") {
        return(termination_type)
      }
    }
  }
  # If all went smoothly we reach here
  "Normal finish"
}

# There are a few environments we don't want to recurse into
is_printable_env <- function(x) {
  is_environment(x) &&
    !(identical(x, rlang::global_env()) ||
      identical(x, rlang::empty_env()) ||
      identical(x, rlang::base_env()) ||
      rlang::is_namespace(x))
}

#' Build element or node label in tree
#'
#' These methods control how the value of a given node is printed. New methods
#' can be added if support is needed for a novel class
#'
#' @inheritParams tree
#' @param opts A list of options that directly mirrors the named arguments of
#'   [tree]. E.g. `list(val_printer = crayon::red)` is equivalent to
#'   `tree(..., val_printer = crayon::red)`.
#'
#' @export
tree_label <- function(x, opts) {
  UseMethod("tree_label")
}

#' @export
tree_label.function <- function(x, opts) {
  func_args <- collapse_and_truncate_vec(methods::formalArgs(x), 5)
  crayon::italic(paste0("function(", func_args, ")"))
}

#' @export
tree_label.environment <- function(x, opts) {
  format.default(x)
}

#' @export
tree_label.NULL <- function(x, opts) {
  "<NULL>"
}

#' @export
tree_label.character <- function(x, opts) {
  # Get rid of new-line so they don't break tree flow
  if (opts$remove_newlines) {
    x <- gsub("\\n", replacement = "\u21B5", x = x, perl = TRUE)
  }

  # Shorten strings if needed
  max_standalone_length <- 35
  max_vec_length <- 15
  max_length <- if (length(x) == 1) max_standalone_length else max_vec_length
  x <- truncate_string(x, max_length)

  tree_label.default(paste0("\"", x, "\""), opts)
}


#' @export
tree_label.default <- function(x, opts) {
  if (rlang::is_atomic(x)) {
    opts$val_printer(collapse_and_truncate_vec(x, opts$max_vec_len))
  } else if (rlang::is_function(x)) {
    # Lots of times function-like functions don't actually trigger the s3 method
    # for function because they dont have function in their class-list. This
    # catches those.
    tree_label.function(x, opts)
  } else if (rlang::is_environment(x)) {
    # Environments also tend to have the same trouble as functions. For instance
    # the srcobject attached to a function's attributes is an environment but
    # doesn't report as one to s3.
    tree_label.environment(x, opts)
  } else if (rlang::is_expression(x) || rlang::is_formula(x)) {
    paste0(label_class(x, opts), " ", crayon::italic(deparse(x)))
  } else {
    # The "base-case" is simply a list-like object.
    label_class(x, opts)
  }
}


collapse_and_truncate_vec <- function(vec, max_length) {
  vec <- as.character(vec)
  too_long <- length(vec) > max_length
  if (too_long) {
    vec <- utils::head(vec, max_length)
    vec <- c(vec, "...")
  }
  paste0(vec, collapse = ", ")
}

truncate_string <- function(char_vec, max_length) {
  ifelse(
    nchar(char_vec) > max_length,
    # Since we add an elipses we need to take a bit more than the max length
    # off. The gsub adds elipses but also makes sure we dont awkwardly end on
    # a space.
    gsub(
      x = substr(char_vec, start = 1, max_length - 3),
      pattern = "\\s*$",
      replacement = "...",
      perl = TRUE
    ),
    char_vec
  )
}

make_type_abrev <- function(x, omit_scalars) {
  if (!rlang::is_atomic(x) || (rlang::is_scalar_atomic(x) && omit_scalars)) {
    return("")
  }

  type_abrev <- switch(
    typeof(x),
    logical = "lgl",
    integer = "int",
    double = "dbl",
    character = "chr",
    complex = "cpl",
    expression = "expr",
    raw = "raw",
    "unknown"
  )

  paste0("<", type_abrev, " [", format(length(x), big.mark = ","), "]>")
}


# Inspired by waldo:::friendly_type_of(). Prints the class name and hierarchy
# encased in angle brackets along with a prefix that tells you what OO system
# the object belongs to (if it does.)
label_class <- function(x, opts) {
  if (is_missing(x)) {
    return("absent")
  }
  oo_prefix <- ""

  class_list <- if (!is.object(x)) {
    typeof(x)
  } else if (isS4(x)) {
    oo_prefix <- "S4"
    methods::is(x)
  } else if (inherits(x, "R6")) {
    oo_prefix <- "R6"
    setdiff(class(x), "R6")
  } else {
    oo_prefix <- "S3"
    class(x)
  }

  opts$class_printer(
    paste0(oo_prefix, "<", paste(class_list, collapse = "/"), ">")
  )
}

test_that("Array-like indices can be shown or hidden", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    tree(list(a = "a", "b", "c"), index_unnamed = TRUE)
  })
  expect_snapshot({
    tree(list(a = "a", "b", "c"), index_unnamed = FALSE)
  })
})

test_that("Atomic arrays have sensible defaults w/ truncation for longer than 10-elements", {
  testthat::skip_on_os("windows")

  expect_snapshot(
    tree(
      list(
        name = "vectored list",
        num_vec = 1:10,
        char_vec = letters
      )
    )
  )

  expect_snapshot(
    tree(
      list(
        name = "vectored list",
        num_vec = 1:10,
        char_vec = letters
      ),
      hide_scalar_types = FALSE
    )
  )
})

test_that("Large and multiline strings are handled gracefully", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    long_strings <- list(
      "normal string" = "first element",
      "really long string" = paste(rep(letters, 4), collapse = ""),
      "vec of long strings" = c(
        "a long\nand multi\nline string element",
        "a fine length",
        "another long\nand also multi\nline string element"
      )
    )

    # No truncation of first string
    # Really long single string is truncated and elipsesed
    # Short string inside vector with long strings is not truncated
    tree(long_strings)

    # Newline removal can be disabled
    tree(long_strings, remove_newlines = FALSE)
  })
})

test_that("Max depth and length can be enforced", {
  # This test also disables the unicode printing so it can be run on windows
  # platforms
  old_opts <- options("lobstr.fancy.tree" = FALSE)
  on.exit(options(old_opts))
  expect_snapshot({
    deep_list <- list(
      list(
        id = "b",
        val = 1,
        children = list(
          list(id = "b1", val = 2.5),
          list(
            id = "b2",
            val = 8,
            children = list(
              list(id = "b21", val = 4)
            )
          )
        )
      ),
      list(id = "a", val = 2)
    )
    tree(deep_list, max_depth = 1)
    tree(deep_list, max_depth = 2)
    tree(deep_list, max_depth = 3)

    tree(deep_list, max_length = 0)
    tree(deep_list, max_length = 2)
    tree(deep_list, max_depth = 1, max_length = 4)
  })
})

test_that("Missing values are caught and printed properly", {
  testthat::skip_on_os("windows")

  expect_snapshot(
    tree(
      list(
        "null-element" = NULL,
        "NA-element" = NA
      )
    )
  )
})

test_that("non-named elements in named list", {
  testthat::skip_on_os("windows")

  expect_snapshot(
    tree(list("a" = 1, "el w/o id"))
  )
})

test_that("Attributes are properly displayed as special children nodes", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    list_w_attrs <- structure(
      list(
        structure(
          list(id = "a", val = 2),
          level = 2,
          name = "first child"
        ),
        structure(
          list(
            id = "b",
            val = 1,
            children = list(
              list(id = "b1", val = 2.5)
            )
          ),
          level = 2,
          name = "second child",
          class = "custom-class"
        ),
        level = "1",
        name = "root"
      )
    )

    # Shows attributes
    tree(list_w_attrs, show_attributes = TRUE)

    # Hides attributes (default)
    tree(list_w_attrs, show_attributes = FALSE)
  })
})

test_that("Can optionally recurse into environments", {
  testthat::skip_on_os("windows")

  # Wrapped in a local to avoid different environment setup for code running in
  # test_that instead of interactively
  # Can't use snapshots here because environment address change on each run
  env_printing <- capture.output(
    local(
      {
        ea <- rlang::env(d = 4, e = 5)
        tree(rlang::env(ea, a = 1, b = 2, c = 3))
      },
      envir = rlang::global_env()
    )
  )

  # Seven total nodes should be printed
  expect_equal(
    length(env_printing),
    4
  )

  # Printed only the names we expected
  expect_equal(
    mean(
      grepl(
        pattern = "(environment|a|b|c):",
        env_printing
      )
    ),
    1
  )

  # Should only print two environment nodes (aka didn't escape past global env)
  expect_equal(
    sum(grepl(pattern = "<environment:", env_printing, fixed = TRUE)),
    1
  )
})

test_that("Function arguments get printed", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    tree(
      list(
        no_args = function() {},
        few_args = function(a, b, c) {},
        lots_of_args = function(d, e, f, g, h, i, j, k, l, m, n, o, p) {}
      )
    )
  })
})

test_that("Handles expressions", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    tree(
      list(
        a = quote(a),
        b = quote(a + 1),
        c = y ~ mx + b
      )
    )
  })
})

test_that("Hidden lists dont cause infinite recursion", {
  testthat::skip_on_os("windows")

  expect_snapshot({
    tree(package_version("1.2.3"))
  })
})

#' Call stack tree
#'
#' Shows the relationship between calls on the stack. This function
#' combines the results of [sys.calls()] and [sys.parents()] yielding a display
#' that shows how frames on the call stack are related.
#'
#' @export
#' @examples
#' # If all evaluation is eager, you get a single tree
#' f <- function() g()
#' g <- function() h()
#' h <- function() cst()
#' f()
#'
#' # You get multiple trees with delayed evaluation
#' try(f())
#'
#' # Pay attention to the first element of each subtree: each
#' # evaluates the outermost call
#' f <- function(x) g(x)
#' g <- function(x) h(x)
#' h <- function(x) x
#' try(f(cst()))
#'
#' # With a little ingenuity you can use it to see how NSE
#' # functions work in base R
#' with(mtcars, {cst(); invisible()})
#' invisible(subset(mtcars, {cst(); cyl == 0}))
#'
#' # You can also get unusual trees by evaluating in frames
#' # higher up the call stack
#' f <- function() g()
#' g <- function() h()
#' h <- function() eval(quote(cst()), parent.frame(2))
#' f()
cst <- function() {
  x <- rlang::trace_back()
  print(x, simplify = "none")
  invisible()
}

#' Find memory location of objects and their children.
#'
#' `obj_addr()` gives the address of the value that `x` points to;
#' `obj_addrs()` gives the address of the components the list,
#' environment, and character vector `x` point to.
#'
#' `obj_addr()` has been written in such away that it avoids taking
#' references to an object.
#'
#' @param x An object
#' @export
#' @examples
#' # R creates copies lazily
#' x <- 1:10
#' y <- x
#' obj_addr(x) == obj_addr(y)
#'
#' y[1] <- 2L
#' obj_addr(x) == obj_addr(y)
#'
#' y <- runif(10)
#' obj_addr(y)
#' z <- list(y, y)
#' obj_addrs(z)
#'
#' y[2] <- 1.0
#' obj_addrs(z)
#' obj_addr(y)
#'
#' # The address of an object is different every time you create it:
#' obj_addr(1:10)
#' obj_addr(1:10)
#' obj_addr(1:10)
obj_addr <- function(x) {
  x <- enquo(x)
  addr <- obj_addr_(quo_get_expr(x), quo_get_env(x))

  if (is_testing()) {
    test_addr_get(addr)
  } else {
    addr
  }
}

#' @export
#' @rdname obj_addr
obj_addrs <- function(x) {
  addrs <- obj_addrs_(x)

  if (is_testing()) {
    vapply(addrs, test_addr_get, character(1), USE.NAMES = FALSE)
  } else {
    addrs
  }
}


test_addr <- child_env(emptyenv(), "__next_id" = 1)

test_addr_get <- function(addr) {
  if (env_has(test_addr, addr)) {
    addr <- env_get(test_addr, addr)
  } else {
    addr <- obj_id(test_addr, addr)
  }
  sprintf("0x%03i", addr)
}

test_addr_reset <- function() {
  env_poke(test_addr, "__next_id", 1)
}

test_that("address of expression varies", {
  a <- obj_addr(1:10)
  b <- obj_addr(1:10)
  expect_false(identical(a, b))
})

test_that("address of variable is constant", {
  x <- 1:10
  expect_equal(obj_addr(x), obj_addr(x))
})

test_that("address flows through function wrappers", {
  x <- 1:10
  f <- function(x) obj_addr(x)
  g <- function(y) f(y)
  h <- function(z) g(z)

  address <- obj_addr(x)
  expect_equal(f(x), address)
  expect_equal(g(x), address)
  expect_equal(h(x), address)
})


# addresses ---------------------------------------------------------------

test_that("can find addresses of list elements", {
  x <- 1:3
  y <- 1:3
  addr <- c(obj_addr(x), obj_addr(y))

  l <- list(x, y)
  expect_equal(obj_addrs(l), addr)
})

test_that("can find addresses of environment elements", {
  x <- 1:3
  y <- 1:3
  addr <- c(obj_addr(x), obj_addr(y))

  e1 <- new.env(hash = TRUE)
  e1$x <- x
  e1$y <- y
  expect_setequal(obj_addrs(e1), addr)

  e2 <- new.env(hash = FALSE)
  e2$x <- x
  e2$y <- y
  expect_setequal(obj_addrs(e2), addr)
})

test_that("address of character vectors points to global string pool", {
  addr <- obj_addrs(c("a", "a", "a"))
  expect_equal(addr[[1]], addr[[2]])
})

test_that("addresses of other elements throws errors", {
  expect_error(obj_addrs(1:10), "must be a list")
})

test_that("basic list display", {
  skip_on_os("windows")

  test_addr_reset()
  expect_snapshot({
    x <- 1:10
    y <- list(x, x)

    ref(
      x,
      list(),
      list(x, x, x),
      list(a = x, b = x),
      letters
    )
  })
})

test_that("basic environment display", {
  skip_on_os("windows")

  test_addr_reset()
  expect_snapshot({
    e <- env(a = 1:10)
    e$b <- e$a
    e$c <- e
    ref(e)
  })
})

test_that("environment shows objects beginning with .", {
  skip_on_os("windows")

  test_addr_reset()
  expect_snapshot({
    e <- env(. = 1:10)
    ref(e)
  })
})


test_that("can display ref to global string pool on request", {
  skip_on_os("windows")

  test_addr_reset()
  expect_snapshot({
    ref(c("string", "string", "new string"), character = TRUE)
  })
})

test_that("custom methods are never called (#30)", {
  # `[[.numeric_number` causes infinite recursion
  expect_error(ref(package_version("1.1.1")), NA)

  e <- env(a = 1:10)
  e$b <- e$a
  e$c <- e

  # `as.list.data.frame`(<environemnt>, ...) fails
  class(e) <- "data.frame"

  expect_error(ref(e), NA)
})

#' Display tree of source references
#'
#' View source reference metadata attached to R objects in a tree structure.
#' Shows source file information, line/column locations, and lines of source code.
#'
#' @param x An R object with source references. Can be:
#'   - A `srcref` object
#'   - A list of `srcref` objects
#'   - A expression vector with attached source references
#'   - An evaluated closure with attached source references
#'   - A quoted call with attached source references
#' @param max_depth Maximum depth to traverse nested structures (default 5)
#' @param max_length Maximum number of srcref nodes to display (default 100)
#' @param ... Additional arguments passed to [tree()]
#'
#' @return Returns a structured list containing the source reference
#'   information. Print it to view the formatted tree.
#'
#' @section Overview:
#'
#' Source references are made of two kinds of objects:
#' - `srcref` objects, which contain information about a specific
#'   location within the source file, such as the line and column numbers.
#' - `srcfile` objects, which contain metadata about the source file
#'   such as its name, path, and encoding.
#'
#'
#' ## Where and when are source references created?
#'
#' Ultimately the R parser creates source references. The main two entry points
#' to the parser are:
#' - The R function `parse()`.
#' - The frontend hook `ReadConsole`, which powers the console input parser in
#'   the R CLI and in IDEs. This C-level parser can also be accessed from C code
#'   via `R_ParseVector()`.
#'
#' In principle, anything that calls `parse()` may create source references, but
#' here are the important direct and indirect callers:
#' - `source()` and `sys.source()` which parse and evaluate code.
#' - `loadNamespace()` calls `sys.source()` when loading a _source_ package:
#'    <https://github.com/r-devel/r-svn/blob/acd196be/src/library/base/R/namespace.R#L573>.
#' - `R CMD install` creates a lazy-load database from a source package.
#'   The first step is to call `loadNamespace()`:
#'   <https://github.com/r-devel/r-svn/blob/acd196be/src/library/tools/R/makeLazyLoad.R#L32>
#'
#' By default source references are not created but can be enabled by:
#'
#' - Passing `keep.source = TRUE` explicitly to `parse()`, `source()`,
#'   `sys.source()`, or `loadNamespace()`.
#' - Setting `options(keep.source = TRUE)`. This affects the default arguments
#'   of the aforementioned functions, as well as the console input parser.
#'   In interactive sessions, `keep.source` is set to `TRUE` by default:
#'   <https://github.com/r-devel/r-svn/blob/3a4745af/src/library/profile/Common.R#L26>.
#' - Setting `options(keep.source.pkgs = TRUE)`. This affects loading a package
#'   from source, and installing a package from source.
#'
#'
#' ## `srcref` objects
#'
#' `srcref` objects are compact integer vectors describing a character range
#' in a source. It records start/end lines and byte/column positions and,
#' optionally, the parsed-line numbers if `#line` directives were used.
#'
#' Lengths of 4, 6, or 8 are allowed:
#' - 4: basic (first_line, first_byte, last_line, last_byte). Byte positions
#'   are within the line.
#' - 6: adds columns in Unicode codepoints (first_col, last_col)
#' - 8: adds parsed-line numbers (first_parsed, last_parsed)
#'
#' The "column" information does not represent grapheme clusters, but Unicode
#' codepoints. The column cursor is incremented at every UTF-8 lead byte and
#' there is no support for encodings other than UTF-8.
#'
#' The srcref columns are right-boundary positions, meaning that for an
#' expression starting at the start of a line, the column will be 1.
#' `wholeSrcref` (see below) on the other hand starts at 0, before the first
#' character. It might also end 1 character after the last srcref column.
#'
#' They are attached as attributes (e.g. `attr(x, "srcref")` or `attr(x,
#' "wholeSrcref")`), possibly wrapped in a list, to the following objects:
#'
#' - Expression vectors returned by `parse()` (wrapped in a list)
#' - Quoted function calls (unwrapped)
#' - Quoted `{` calls (wrapped in a list). This is crucial for debugging: when R
#'   steps through brace lists, the srcref for the current expression is saved to
#'   a global variable (`R_Srcref`) so the IDE knows exactly where execution is
#'   paused. See: <https://github.com/r-devel/r-svn/blob/fa0b47c5/src/main/eval.c#L2986>.
#' - Evaluated closures (unwrapped)
#'
#' They have a `srcfile` attribute that points to the source file.
#'
#' Methods:
#' - `as.character()`: Retrieves relevant source lines from the `srcfile`
#'   reference.
#'
#'
#' ### `wholeSrcref` attributes
#'
#' These are `srcref` objects stored in the `wholeSrcref` attributes of:
#'
#' - Expression vectors returned by `parse()`, which seems to be the intended
#'   usage.
#' - `{` calls, which seems unintended.
#'
#' For expression vectors, the `wholeSrcref` spans from the first position
#' to the last position and represents the entire document. For braces, they
#' span from the first position to the location of the closing brace. There is
#' no way to know the location of the opening brace without reparsing, which
#' seems odd. It's probably an overlook from `xxexprlist()` calling
#' `attachSrcrefs()` in
#' <https://github.com/r-devel/r-svn/blob/52affc16/src/main/gram.y#L1380>. That
#' function is also called at the end of parsing, where it's intended for the
#' `wholeSrcref` attribute to be attached.
#'
#' For evaluated closures, the `wholeSrcref` attribute on the body has the same
#' unreliable start positions as `{` nodes.
#'
#'
#' ## `srcfile` objects
#'
#' `srcfile` objects are environments representing information about a
#' source file that a source reference points to. They typically refer to
#' a file on disk and store the filename, working directory, a timestamp,
#' and encoding information.
#'
#' While it is possible to create bare `srcfile` objects, specialized subclasses
#' are much more common.
#'
#'
#' ### `srcfile`
#'
#' A bare `srcfile` object does not contain any data apart from the file path.
#' It lazily loads lines from the file on disk, without any caching.
#'
#' Fields common to all `srcfile` objects:
#'
#' - `filename`: The filename of the source file. If relative, the path is
#'   resolved against `wd`.
#'
#' - `wd`: The working directory (`getwd()`) at the time the srcfile was created,
#'   generally at the time of parsing).
#'
#' - `timestamp`: The timestamp of the source file. Retrieved from `filename`
#'   with `file.mtime()`.
#'
#' - `encoding`: The encoding of the source file.
#'
#' - `Enc`: The encoding of output lines. Used by `getSrcLines()`, which
#'   calls `iconv()` when `Enc` does not match `encoding`.
#'
#' - `parseData` (optional): Parser information saved when `keep.source.data` is
#'    set to `TRUE`.
#'
#' Implementations:
#' - `print()` and `summary()` to print information about the source file.
#' - `open()` and `close()` to access the underlying file as a connection.
#'
#' Helpers:
#' - `getSrcLines()`: Retrieves source lines from a `srcfile`.
#'
#'
#' ### `srcfilecopy`
#'
#' A `srcfilecopy` stores the actual source lines in memory in `$lines`.
#' `srcfilecopy` is useful when the original file may change or does not
#' exist, because it preserves the exact text used by the parser.
#'
#' This type of srcfile is the most common. It's created by:
#'
#' - The R-level `parse()` function when `text` is supplied:
#'
#'   ```r
#'   # Creates a `"<text>"` non-file `srcfilecopy`
#'   parse(text = "...", keep.source = TRUE)
#'   ```
#'
#' - The console's input parser when `getOption("keep.source")` is `TRUE`.
#'
#' - `sys.source()` when `keep.source = TRUE`:
#'
#'   ```r
#'   sys.source(file, keep.source = TRUE)
#'   ```
#'
#'    The `srcfilecopy` object is timestamped with the file's last modification time.
#'    <https://github.com/r-devel/r-svn/blob/52affc16/src/library/base/R/source.R#L273-L276>
#'
#' Fields:
#'
#' - `filename`: The filename of the source file. If `isFile` is `FALSE`,
#'   the field is non meaningful. For instance `parse(text = )` sets it to
#'   `"<text>"`, and the console input parser sets it to `""`.
#'
#' - `isFile`: A logical indicating whether the source file exists.
#'
#' - `fixedNewlines`: If `TRUE`, `lines` is a character vector of lines with
#'   no embedded `\n` characters. The `getSrcLines()` helper regularises `lines`
#'   in this way and sets `fixedNewlines` to `TRUE`.
#'
#'
#' Note that the C-level parser (used directly mainly when parsing console input)
#' does not call the R-level constructor and only instantiates the `filename`
#' (set to `""`) and `lines` fields.
#'
#' ### `srcfilealias`
#'
#' This object wraps an existing `srcfile` object (stored in `original`).  It
#' allows exposing a different `filename` while delegating the open/close/get
#' lines operations to the `srcfile` stored in `original`.
#'
#' The typical way aliases are created is via `#line *line* *filename*`
#' directives where the optional `*filename*` argument is supplied. These
#' directives remap the srcref and srcfile of parsed code to a different
#' location, for example from a temporary file or generated file to the original
#' location on disk.
#'
#' Created by `install.packages()` when installing a _source_ package with `keep.source.pkgs` set to `TRUE` (see
#' <https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/install.R#L545>), but
#' [only when](https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/admin.R#L308):
#'
#' - `Encoding` was supplied in `DESCRIPTION`
#' - The system locale is not "C" or "POSIX".
#'
#' The source files are converted to the encoding of the system locale, then
#' collated in a single source file with `#line` directives mapping them to their
#' original file names (with full paths):
#' <https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/admin.R#L342>.
#'
#' Note that the `filename` of the `original` srcfile incorrectly points to the
#' package path in the install destination.
#'
#'
#' Fields:
#'
#' - `filename`: The virtual file name (or full path) of the parsed code.
#' - `original`: The actual `srcfile` the code was parsed from.
#'
#' @seealso
#' - [srcfile()]: Base documentation for `srcref` and `srcfile` objects.
#' - [getParseData()]: Parse information stored when `keep.source.data` is `TRUE`.
#' - Source References (R Journal): <https://journal.r-project.org/articles/RJ-2010-010/>
#'
#' @export
#' @family object inspectors
src <- function(
  x,
  max_depth = 5L,
  max_length = 100L,
  ...
) {
  seen_srcfiles <- new.env(parent = emptyenv())
  seen_srcfiles$.counter <- 0L

  result <- src_extract(x, seen_srcfiles)
  if (is.null(result)) {
    return(invisible(NULL))
  }

  # Ensure result has proper type for tree display
  if (is.null(attr(result, "srcref_type"))) {
    result <- as_srcref_tree(result, from = x)
  }

  structure(
    result,
    max_depth = max_depth,
    max_length = max_length,
    tree_args = list(...),
    class = c("lobstr_srcref", class(result))
  )
}

#' @export
print.lobstr_srcref <- function(x, ...) {
  max_depth <- attr(x, "max_depth") %||% 5L
  max_length <- attr(x, "max_length") %||% 100L
  tree_args <- attr(x, "tree_args") %||% list()

  # Strip attributes before printing
  attr(x, "max_depth") <- NULL
  attr(x, "max_length") <- NULL
  attr(x, "tree_args") <- NULL

  # Defaults for `tree()` arguments that are not directly exposed by `src()`
  tree_args$max_vec_len <- tree_args$max_vec_len %||% 3L

  inject(tree(
    x = x,
    max_depth = max_depth,
    max_length = max_length,
    !!!tree_args
  ))

  invisible(x)
}

#' @export
tree_label.lobstr_srcref <- function(x, opts) {
  type <- attr(x, "srcref_type")

  switch(
    type,
    body = "",
    block = "<{>",
    srcfile = srcfile_label(x),
    paste0("<", type, ">")
  )
}

#' @export
tree_label.lobstr_srcref_location <- function(x, opts) {
  as.character(x)
}

#' @export
tree_label.srcref <- function(x, opts) {
  loc <- srcref_location(x)
  paste0("<srcref: ", loc, ">")
}

#' @export
tree_label.srcfile <- function(x, opts) {
  paste0("<", class(x)[1], ": ", utils::getSrcFilename(x), ">")
}

#' @export
tree_label.lobstr_srcfile_ref <- function(x, opts) {
  paste0("@", as.character(x))
}


# Main extraction logic --------------------------------------------------------

src_extract <- function(x, seen_srcfiles) {
  # Srcref object
  if (inherits(x, "srcref")) {
    return(srcref_node(x, seen_srcfiles))
  }

  # List of srcrefs
  if (
    is.list(x) &&
      length(x) > 0 &&
      all(vapply(x, inherits, logical(1), "srcref"))
  ) {
    return(srcref_list_node(x, seen_srcfiles))
  }

  # Evaluated closures
  if (is_closure(x)) {
    return(function_node(x, seen_srcfiles))
  }

  # Expressions and language objects
  if (is.expression(x) || is.language(x)) {
    return(expr_node(x, seen_srcfiles))
  }

  NULL
}

# Extract standard srcref-related attributes from any object
extract_srcref_attrs <- function(x, seen_srcfiles) {
  attrs <- list()

  if (!is.null(srcref <- attr(x, "srcref"))) {
    attrs$`attr("srcref")` <- srcref_attr_node(
      srcref,
      seen_srcfiles
    )
  }

  if (!is.null(srcfile <- attr(x, "srcfile"))) {
    attrs$`attr("srcfile")` <- srcfile_node(srcfile, seen_srcfiles)
  }

  if (!is.null(whole <- attr(x, "wholeSrcref"))) {
    attrs$`attr("wholeSrcref")` <- srcref_attr_node(whole, seen_srcfiles)
  }

  attrs
}

# A srcref attribute may be a srcref object or a list of srcref objects
srcref_attr_node <- function(srcref, seen_srcfiles) {
  if (inherits(srcref, "srcref")) {
    return(srcref_node(srcref, seen_srcfiles))
  }

  if (is.list(srcref)) {
    return(srcref_list_node(srcref, seen_srcfiles))
  }

  NULL
}

srcref_node <- function(srcref, seen_srcfiles) {
  info <- srcref_info(srcref)
  node <- list(location = info$location)

  if (!is.null(info$bytes)) {
    node$bytes <- info$bytes
  }
  if (!is.null(info$parsed)) {
    node$parsed <- info$parsed
  }

  # Just for completeness but we really don't expect srcref attributes on srcrefs
  attrs <- extract_srcref_attrs(srcref, seen_srcfiles)
  node <- c(node, attrs)

  new_srcref_tree(node, type = "srcref")
}

srcref_list_node <- function(srcref_list, seen_srcfiles) {
  srcrefs <- lapply(srcref_list, srcref_node, seen_srcfiles)
  names(srcrefs) <- paste0("[[", seq_along(srcrefs), "]]")

  attrs <- extract_srcref_attrs(srcref_list, seen_srcfiles)
  node <- c(srcrefs, attrs)

  new_srcref_tree(node, type = "list")
}

function_node <- function(fun, seen_srcfiles) {
  node <- extract_srcref_attrs(fun, seen_srcfiles)
  body <- src_extract(body(fun), seen_srcfiles)

  if (!is.null(body)) {
    node$`body()` <- as_srcref_tree(body, from = body(fun))
  }

  if (length(node) == 0) {
    return(NULL)
  }

  new_srcref_tree(node, type = "closure")
}

expr_node <- function(x, seen_srcfiles) {
  attrs <- extract_srcref_attrs(x, seen_srcfiles)
  nested <- extract_nested_srcrefs(x, seen_srcfiles)

  if (length(attrs) > 0) {
    # Node has attributes: wrap with proper type
    node <- c(attrs, nested)
    return(new_srcref_tree(node, type = node_type(x)))
  }

  # No attributes: return bare list for path collapsing, or NULL if empty
  if (length(nested) > 0) {
    nested
  } else {
    NULL
  }
}

extract_nested_srcrefs <- function(x, seen_srcfiles) {
  if (!is_traversable(x)) {
    return(list())
  }

  nested <- list()
  for (i in seq_along(x)) {
    child <- src_extract(x[[i]], seen_srcfiles)

    if (!is.null(child)) {
      nested <- merge_child_result(nested, child, i)
    }
  }

  nested
}

merge_child_result <- function(nested, child, index) {
  path <- paste0("[[", index, "]]")

  if (is_wrapped_node(child)) {
    nested[[path]] <- child
  } else {
    # Collapse paths for bare lists
    for (name in names(child)) {
      nested[[paste0(path, name)]] <- child[[name]]
    }
  }

  nested
}

is_traversable <- function(x) {
  (is.expression(x) || is.call(x)) && length(x) > 0
}

is_wrapped_node <- function(x) {
  !is.null(attr(x, "srcref_type"))
}

node_type <- function(x) {
  if (is.expression(x)) {
    "expression"
  } else if (is.call(x) && length(x) > 0) {
    if (identical(x[[1]], as.symbol("function"))) {
      "quoted_function"
    } else if (identical(x[[1]], as.symbol("{"))) {
      "block"
    } else {
      "language"
    }
  } else {
    "language"
  }
}

as_srcref_tree <- function(data, ..., from) {
  if (is_wrapped_node(data)) {
    data
  } else {
    new_srcref_tree(data, type = node_type(from))
  }
}


# Srcfile handling -------------------------------------------------------------

srcfile_node <- function(srcfile, seen_srcfiles) {
  if (is.null(srcfile)) {
    return(NULL)
  }

  addr <- obj_addr(srcfile)
  srcfile_class <- class(srcfile)[[1]]

  # Check if already seen
  id <- seen_srcfiles[[addr]]
  if (!is_null(id)) {
    return(new_srcfile_ref(id, srcfile_class))
  }

  # First occurrence - assign sequential ID
  seen_srcfiles$.counter <- seen_srcfiles$.counter + 1L
  id <- sprintf("%03d", seen_srcfiles$.counter)
  seen_srcfiles[[addr]] <- id

  info <- as.list.environment(srcfile, all.names = TRUE, sorted = TRUE)

  # Format timestamp for readability
  if (!is.null(info$timestamp)) {
    info$timestamp <- format(info$timestamp)
  }

  # Process nested srcfile objects (e.g., 'original' in srcfilealias)
  if (!is.null(info$original) && inherits(info$original, "srcfile")) {
    info$original <- srcfile_node(info$original, seen_srcfiles)
  }

  # Add source preview for plain srcfiles
  if (!inherits(srcfile, "srcfilecopy") && !is.null(srcref)) {
    snippet <- srcfile_lines(srcfile, srcref)
    if (length(snippet) > 0) {
      info$`lines (from file)` <- snippet
    }
  }

  # Check for srcref attributes even on srcfile objects
  attrs <- extract_srcref_attrs(srcfile, seen_srcfiles)
  info <- c(info, attrs)

  new_srcref_tree(
    info,
    type = "srcfile",
    srcfile_class = srcfile_class %||% "srcfile",
    srcfile_id = id
  )
}

srcfile_lines <- function(srcfile, srcref) {
  if (is.null(srcfile) || !is_srcref(srcref)) {
    return(character(0))
  }

  max_lines <- 3L

  first_line <- srcref[[1]]
  last_line <- min(srcref[[3]], first_line + max_lines - 1L)

  # Try embedded lines first
  lines <- srcfile$lines
  if (!is.null(lines) && length(lines) >= last_line) {
    return(lines[first_line:last_line])
  }

  # Try reading from file
  filename <- srcfile$filename
  directory <- srcfile$wd

  if (!is.null(filename) && !is.null(directory)) {
    filepath <- file.path(directory, filename)

    if (file.exists(filepath)) {
      all_lines <- tryCatch(
        readLines(filepath, warn = FALSE),
        error = function(e) NULL
      )

      if (!is.null(all_lines) && length(all_lines) >= last_line) {
        return(all_lines[first_line:last_line])
      }
    }
  }

  character(0)
}

srcfile_label <- function(x) {
  class <- attr(x, "srcfile_class")
  label <- paste0("<", class, ">")

  id <- attr(x, "srcfile_id")
  if (!is.null(id)) {
    label <- paste0(label, " @", id)
  }

  label
}


# Srcref information extraction ------------------------------------------------

srcref_info <- function(srcref) {
  if (!inherits(srcref, "srcref")) {
    abort("Expected a srcref object")
  }

  len <- length(srcref)
  if (!len %in% c(4, 6, 8)) {
    abort(sprintf("Unexpected srcref length: %d", len))
  }

  first_line <- srcref[[1]]
  first_byte <- srcref[[2]]
  last_line <- srcref[[3]]
  last_byte <- srcref[[4]]
  first_col <- if (len >= 6) srcref[[5]] else first_byte
  last_col <- if (len >= 6) srcref[[6]] else last_byte
  first_parsed <- if (len == 8) srcref[[7]] else first_line
  last_parsed <- if (len == 8) srcref[[8]] else last_line

  info <- list(
    location = new_srcref_location(srcref_location(srcref))
  )

  # Add byte info if different from columns
  if (first_byte != first_col || last_byte != last_col) {
    info$bytes <- sprintf("%d - %d", first_byte, last_byte)
  }

  # Add parsed info if different from actual lines
  if (first_parsed != first_line || last_parsed != last_line) {
    info$parsed <- new_srcref_location(sprintf(
      "%d:%d - %d:%d",
      first_parsed,
      first_col,
      last_parsed,
      last_col
    ))
  }

  info
}

srcref_location <- function(x) {
  first_line <- x[[1]]
  last_line <- x[[3]]
  first_col <- if (length(x) >= 6) x[[5]] else x[[2]]
  last_col <- if (length(x) >= 6) x[[6]] else x[[4]]

  sprintf("%d:%d - %d:%d", first_line, first_col, last_line, last_col)
}


# Helper functions -------------------------------------------------------------

has_srcref <- function(x) {
  !is.null(attr(x, "srcref")) ||
    !is.null(attr(x, "wholeSrcref")) ||
    !is.null(attr(x, "srcfile"))
}

new_srcref_tree <- function(x, type = NULL, ..., class = NULL) {
  type <- type %||% attr(x, "srcref_type")
  type <- arg_match(
    type,
    c(
      "block",
      "body",
      "closure",
      "expression",
      "language",
      "list",
      "quoted_function",
      "srcfile",
      "srcref"
    )
  )

  structure(
    x,
    srcref_type = type,
    ...,
    class = c(class, "lobstr_srcref")
  )
}

new_srcref_location <- function(x) {
  structure(x, class = c("lobstr_srcref_location", "character"))
}

new_srcfile_ref <- function(id, srcfile_class = "srcfile") {
  structure(
    id,
    srcfile_class = srcfile_class,
    class = "lobstr_srcfile_ref"
  )
}

is_srcref <- function(x) {
  is.integer(x) && inherits(x, "srcref") && length(x) %in% c(4L, 6L, 8L)
}

# Test: Closures (evaluated functions) ------------------------------------------

if (utils::packageVersion("base") < "4.2.0") {
  # Tree characters are ASCII on old Windows R
  skip_on_os("windows")
}

test_that("src() shows closure with srcref and wholeSrcref", {
  expect_snapshot(transform = scrub_src_transform, {
    f <- simple_function_with_srcref()
    src(f)
  })
})

test_that("src() shows multi-statement function", {
  expect_snapshot(transform = scrub_src_transform, {
    f <- multi_statement_function_with_srcref()
    src(f)
  })
})

# Test: Quoted functions --------------------------------------------------------

test_that("src() shows quoted function with nested body", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("x <- quote(function() {})")
    src(x)
  })
})

test_that("src() shows quoted function body directly", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("x <- quote(function() {})")
    src(x[[3]])
  })
})

test_that("src() shows quoted function with arguments", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("x <- quote(function(a, b) {})")
    src(x)
  })
})

test_that("src() shows srcref with parsed field when positions differ", {
  expect_snapshot(transform = scrub_src_transform, {
    # Create a synthetic 8-element srcref where parsed positions differ
    # Format: c(first_line, first_byte, last_line, last_byte,
    #           first_col, last_col, first_parsed, last_parsed)
    # This simulates a case where R's parser reports different positions
    # than the actual source locations (e.g., due to string continuations)
    srcfile <- srcfilecopy(
      "test.R",
      c(
        "x <- function() {",
        "  # A long comment that spans",
        "  # multiple lines",
        "  y <- 1",
        "}"
      )
    )

    synthetic_srcref <- structure(
      c(2L, 3L, 4L, 8L, 3L, 8L, 1L, 5L),
      class = "srcref",
      srcfile = srcfile
    )

    src(synthetic_srcref)
  })
})

# Test: Expression objects ------------------------------------------------------

test_that("src() shows expression with single element", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "x + 1", keep.source = TRUE)
    src(x)
  })
})

test_that("src() shows expression with multiple elements", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = c("x + 1", "y + 2", "z + 3"), keep.source = TRUE)
    src(x)
  })
})

test_that("src() shows expression with nested block and wholeSrcref", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "{\n  1\n}", keep.source = TRUE)
    src(x)
  })
})

test_that("src() shows nested block element directly", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "{\n  1\n}", keep.source = TRUE)
    src(x[[1]])
  })
})

# Test: Blocks with wholeSrcref -------------------------------------------------

test_that("src() shows block with srcref list and wholeSrcref", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "{\n  a <- 1\n  b <- 2\n}", keep.source = TRUE)
    src(x[[1]])
  })
})

# Test: Single srcref objects ---------------------------------------------------

test_that("src() shows single srcref", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "x + 1", keep.source = TRUE)
    sr <- attr(x, "srcref")[[1]]
    src(sr)
  })
})

# Test: List of srcrefs ---------------------------------------------------------

test_that("src() shows list of srcrefs with count", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = c("x + 1", "y + 2"), keep.source = TRUE)
    sr_list <- attr(x, "srcref")
    src(sr_list)
  })
})

# Test: Srcref lists shown as <list> with [[1]], [[2]] notation ----------------

test_that("src() reveals srcref list structure with index notation", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("x <- quote(function() { 1 })")
    src(x[[3]])
  })
})

test_that("src() handles srcrefs nested in language calls", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "foo({ if (1) bar({ 2 }) })", keep.source = TRUE)
    src(x, max_depth = 10)
  })
})

test_that("src() handles srcrefs nested in function bodies", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("f <- function() foo({ if (1) bar({ 2 }) })")
    src(f, max_depth = 10)
  })
})

# Test: Type labels -------------------------------------------------------------

test_that("src() uses correct type labels", {
  # Closure
  f <- simple_function_with_srcref()
  result_closure <- src(f)
  expect_equal(attr(result_closure, "srcref_type"), "closure")

  # Quoted function
  with_srcref("x <- quote(function() {})")
  result_quoted <- src(x)
  expect_equal(attr(result_quoted, "srcref_type"), "quoted_function")

  # Expression
  expr <- parse(text = "1 + 1", keep.source = TRUE)
  result_expr <- src(expr)
  expect_equal(attr(result_expr, "srcref_type"), "expression")

  # Block
  block <- parse(text = "{1}", keep.source = TRUE)[[1]]
  result_block <- src(block)
  expect_equal(attr(result_block, "srcref_type"), "block")
})

# Test: Srcfile duplication (current behavior - will change in Phase 1) --------

test_that("src() currently shows duplicate srcfile objects", {
  expect_snapshot(transform = scrub_src_transform, {
    # Current behavior: srcfile appears twice (in srcref and wholeSrcref)
    # After Phase 1: should use reference notation like @abc123
    f <- simple_function_with_srcref()
    src(f)
  })
})

test_that("src() shows many duplicate srcfiles in nested expression", {
  expect_snapshot(transform = scrub_src_transform, {
    # Current behavior: same srcfile appears many times
    # After Phase 1: these should be deduplicated
    x <- parse(text = "{\n  1\n  2\n}", keep.source = TRUE)
    src(x)
  })
})

# Test: Edge cases --------------------------------------------------------------

test_that("src() handles empty block", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "{}", keep.source = TRUE)
    src(x[[1]])
  })
})

test_that("src() handles function without arguments", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("f <- function() { NULL }")
    src(f)
  })
})

test_that("src() handles if statement with blocks", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(text = "if (TRUE) { 1 } else { 2 }", keep.source = TRUE)
    src(x)
  })
})

# Test: Parameters --------------------------------------------------------------

test_that("src() respects show_source_lines parameter", {
  expect_snapshot(transform = scrub_src_transform, {
    f <- simple_function_with_srcref()
    src(f)
  })
})

# Test: Complex nested structures -----------------------------------------------

test_that("src() shows expression with multiple nested blocks", {
  expect_snapshot(transform = scrub_src_transform, {
    x <- parse(
      text = "{\n  {\n    1\n  }\n  {\n    2\n  }\n}",
      keep.source = TRUE
    )
    src(x)
  })
})

test_that("src() shows function with nested block in body", {
  expect_snapshot(transform = scrub_src_transform, {
    with_srcref("f <- function(x) {\n  if (x) {\n    1\n  }\n}")
    src(f)
  })
})
# Tests for src() function and helpers

# Helper function tests --------------------------------------------------------

test_that("extract_srcref_info handles 4-element srcrefs", {
  # Create a simple expression with srcref
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]

  # Manually create a 4-element srcref for testing
  # Note: In practice, 4-element srcrefs are rare in modern R
  srcref_4 <- structure(
    c(1L, 1L, 1L, 5L),
    class = "srcref",
    srcfile = attr(srcref, "srcfile")
  )

  info <- srcref_info(srcref_4)

  expect_s3_class(info$location, "lobstr_srcref_location")
  expect_equal(as.character(info$location), "1:1 - 1:5")
})

test_that("extract_srcref_info handles 6-element srcrefs", {
  # Create a 6-element srcref
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref_base <- attr(expr, "srcref")[[1]]

  srcref_6 <- structure(
    c(1L, 1L, 1L, 5L, 1L, 5L),
    class = "srcref",
    srcfile = attr(srcref_base, "srcfile")
  )

  info <- srcref_info(srcref_6)

  expect_s3_class(info$location, "lobstr_srcref_location")
  expect_equal(as.character(info$location), "1:1 - 1:5")
})

test_that("extract_srcref_info handles 8-element srcrefs", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]

  # Most modern srcrefs are 8-element
  info <- srcref_info(srcref)

  expect_s3_class(info$location, "lobstr_srcref_location")
  expect_match(as.character(info$location), "\\d+:\\d+ - \\d+:\\d+")
})

test_that("extract_srcref_info shows encoding details when requested", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]

  info <- srcref_info(srcref)

  expect_true("location" %in% names(info))
})

test_that("extract_srcref_info errors on invalid srcref length", {
  # Create an invalid srcref with wrong number of elements
  bad_srcref <- structure(c(1L, 2L, 3L), class = "srcref")

  expect_error(
    srcref_info(bad_srcref),
    "Unexpected srcref length"
  )
})

test_that("srcfile_node handles srcfilecopy", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcfile <- attr(attr(expr, "srcref")[[1]], "srcfile")
  srcref <- attr(expr, "srcref")[[1]]
  seen_srcfiles <- new.env(parent = emptyenv())

  info <- srcfile_node(srcfile, seen_srcfiles)

  expect_equal(attr(info, "srcfile_class"), class(srcfile)[1])
  expect_type(info$filename, "character")
  expect_type(info$Enc, "character")
})

test_that("srcfile_node handles NULL gracefully", {
  seen_srcfiles <- new.env(parent = emptyenv())
  info <- srcfile_node(NULL, seen_srcfiles)
  expect_null(info)
})

test_that("srcfile_lines extracts from srcfilecopy", {
  code <- c("x <- 1", "y <- 2", "z <- 3")
  expr <- parse(text = code, keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]
  srcfile <- attr(srcref, "srcfile")

  snippet <- srcfile_lines(srcfile, srcref)

  expect_type(snippet, "character")
  expect_true(length(snippet) >= 1)
})

test_that("srcfile_lines respects max_lines", {
  code <- c("x <- 1", "y <- 2", "z <- 3", "a <- 4", "b <- 5")
  expr <- parse(text = paste(code, collapse = "\n"), keep.source = TRUE)

  srcfile <- attr(attr(expr, "srcref")[[1]], "srcfile")
  srcref <- structure(
    c(1L, 1L, 5L, 10L, 1L, 10L, 1L, 5L),
    class = "srcref",
    srcfile = srcfile
  )

  snippet <- srcfile_lines(srcfile, srcref)

  expect_type(snippet, "character")
  expect_lte(length(snippet), 3)
})

test_that("srcref_location works correctly", {
  srcref <- structure(
    c(1L, 5L, 3L, 20L, 5L, 20L, 1L, 3L),
    class = "srcref"
  )
  loc <- srcref_location(srcref)
  expect_equal(loc, "1:5 - 3:20")
})

# Integration tests for src() --------------------------------------------------

test_that("src works with functions with source references", {
  fun <- simple_function_with_srcref()

  result <- src(fun)

  expect_type(result, "list")
  expect_equal(attr(result, "srcref_type"), "closure")
})

test_that("src works with single srcref objects", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]

  result <- src(srcref)

  expect_type(result, "list")
  expect_equal(attr(result, "srcref_type"), "srcref")
  expect_true("location" %in% names(result))
})

test_that("src works with list of srcrefs", {
  expr <- parse(text = c("x + 1", "y + 2"), keep.source = TRUE)
  srcref_list <- attr(expr, "srcref")

  result <- src(srcref_list)

  expect_type(result, "list")
  expect_equal(attr(result, "srcref_type"), "list")
})

test_that("src works with expressions", {
  expr <- parse(text = "x + 1", keep.source = TRUE)

  result <- src(expr)

  expect_type(result, "list")
})

test_that("src works for objects without srcrefs", {
  fun <- function(x) x + 1
  attr(fun, "srcref") <- NULL
  expect_null(src(fun))
  expect_null(src(new.env()))
  expect_null(src(list()))
  expect_null(src(sum))
})

test_that("src respects max_lines_preview parameter", {
  fun <- multi_statement_function_with_srcref()

  result <- src(fun, max_lines_preview = 1)

  expect_type(result, "list")
  expect_equal(attr(result, "srcref_type"), "closure")
})

test_that("src returns structure and print method works", {
  fun <- simple_function_with_srcref()

  # src() returns visibly (with S3 class)
  result <- src(fun)
  expect_s3_class(result, "lobstr_srcref")

  # print method returns invisibly and outputs to console
  expect_output(
    expect_invisible(print(result)),
    "<closure>"
  )
})

# S3 method tests --------------------------------------------------------------

test_that("tree_label.srcref formats correctly", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]

  # Call the method directly since srcref has proper class
  label <- tree_label.srcref(srcref, list())

  expect_type(label, "character")
  expect_match(label, "<srcref:")
  expect_match(label, "\\d+:\\d+ - \\d+:\\d+")
})

test_that("tree_label.srcfile formats correctly", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcfile <- attr(attr(expr, "srcref")[[1]], "srcfile")

  # Call the method directly since srcfile is an environment with class attribute
  label <- tree_label.srcfile(srcfile, list())

  expect_type(label, "character")
  expect_match(label, "<srcfile")
})

# Edge case tests --------------------------------------------------------------

test_that("src handles functions with wholeSrcref on body", {
  fun <- simple_function_with_srcref()

  # Verify wholeSrcref exists on body
  expect_false(is.null(attr(body(fun), "wholeSrcref")))

  result <- src(fun)
  expect_type(result, "list")

  # Should have body() node in the result
  expect_true("body()" %in% names(result))
  expect_s3_class(result$`body()`, "lobstr_srcref")

  # wholeSrcref should be nested under body()
  expect_true("attr(\"wholeSrcref\")" %in% names(result$`body()`))
  expect_s3_class(result$`body()`$`attr("wholeSrcref")`, "lobstr_srcref")
})

test_that("src handles functions with only wholeSrcref (no srcref attr)", {
  fun <- simple_function_with_srcref()

  # Remove srcref attribute from function, keep only wholeSrcref on body
  attr(fun, "srcref") <- NULL

  result <- src(fun)
  expect_type(result, "list")

  # Should still have body() node with wholeSrcref from body
  expect_true("body()" %in% names(result))
  expect_true("attr(\"wholeSrcref\")" %in% names(result$`body()`))
})

test_that("srcfile_lines handles missing files gracefully", {
  expr <- parse(text = "x + 1", keep.source = TRUE)
  srcref <- attr(expr, "srcref")[[1]]
  srcfile <- attr(srcref, "srcfile")

  # Point to a non-existent file
  srcfile$filename <- "nonexistent_file.R"

  snippet <- srcfile_lines(srcfile, srcref)

  expect_type(snippet, "character")
})

# Srcfile deduplication tests --------------------------------------------------

test_that("srcfile deduplication - basic case with srcref and wholeSrcref", {
  # Parse code that creates a function with both srcref and wholeSrcref
  code <- parse(text = "f <- function(x) { x + 1 }", keep.source = TRUE)
  f <- eval(code[[1]])

  result <- src(f)

  # Extract the srcfile from the first occurrence
  first_srcfile <- result$`attr("srcref")`$`attr("srcfile")`

  # Extract the ID from the first occurrence
  id <- attr(first_srcfile, "srcfile_id")
  expect_type(id, "character")
  expect_true(nchar(id) >= 1) # Should be hex ID (up to 6 chars)
  expect_true(nchar(id) <= 6)

  # Check that wholeSrcref uses a reference
  whole_srcfile <- result$`body()`$`attr("wholeSrcref")`$`attr("srcfile")`
  expect_s3_class(whole_srcfile, "lobstr_srcfile_ref")
  expect_equal(as.character(whole_srcfile), id)
})

test_that("srcfile deduplication - multiple statement srcrefs share one srcfile", {
  # Create a function with multiple statements
  code <- parse(
    text = "f <- function(x) { a <- x + 1; b <- a * 2; b }",
    keep.source = TRUE
  )
  f <- eval(code[[1]])

  result <- src(f)

  # Get the ID from the first occurrence
  first_srcfile <- result$`attr("srcref")`$`attr("srcfile")`
  id <- attr(first_srcfile, "srcfile_id")

  # Check that all statement srcrefs use references
  stmt_list <- result$`body()`$`attr("srcref")`

  for (i in seq_along(stmt_list)) {
    stmt_name <- paste0("[[", i, "]]")
    stmt_srcfile <- stmt_list[[stmt_name]]$`attr("srcfile")`

    # Should be a reference
    expect_s3_class(stmt_srcfile, "lobstr_srcfile_ref")
    expect_equal(as.character(stmt_srcfile), id)
  }
})

test_that("srcfile deduplication - IDs are stable within a single src() call", {
  # Parse the same code twice to get two different srcfile objects
  code1 <- parse(text = "f <- function(x) { x + 1 }", keep.source = TRUE)
  code2 <- parse(text = "g <- function(y) { y * 2 }", keep.source = TRUE)

  f <- eval(code1[[1]])
  g <- eval(code2[[1]])

  # Call src() on first function
  result_f <- src(f)
  id_f <- attr(result_f$`attr("srcref")`$`attr("srcfile")`, "srcfile_id")

  # Call src() on second function (different call, different seen_srcfiles)
  result_g <- src(g)
  id_g <- attr(result_g$`attr("srcref")`$`attr("srcfile")`, "srcfile_id")

  # IDs are sequential and start fresh for each src() call
  expect_type(id_f, "character")
  expect_type(id_g, "character")

  # Both should be 3-digit sequential IDs starting at "001"
  expect_equal(id_f, "001")
  expect_equal(id_g, "001")
})

test_that("srcfile deduplication - multiple files means no cross-file deduplication", {
  # Parse two separate code snippets (different srcfile objects)
  code1 <- parse(text = "f <- function(x) { x + 1 }", keep.source = TRUE)
  code2 <- parse(text = "g <- function(y) { y * 2 }", keep.source = TRUE)

  f <- eval(code1[[1]])
  g <- eval(code2[[1]])

  # Get the srcfile addresses
  srcfile_f <- attr(attr(f, "srcref"), "srcfile")
  srcfile_g <- attr(attr(g, "srcref"), "srcfile")

  addr_f <- lobstr::obj_addr(srcfile_f)
  addr_g <- lobstr::obj_addr(srcfile_g)

  # Different srcfiles should have different addresses
  expect_false(addr_f == addr_g)

  # If we call src() on each separately, they each get their own ID
  result_f <- src(f)
  result_g <- src(g)

  id_f <- attr(result_f$`attr("srcref")`$`attr("srcfile")`, "srcfile_id")
  id_g <- attr(result_g$`attr("srcref")`$`attr("srcfile")`, "srcfile_id")

  # IDs are sequential and start fresh for each src() call, so both get "001"
  # This ensures deterministic snapshots
  expect_equal(id_f, "001")
  expect_equal(id_g, "001")
})

test_that("srcfile deduplication - nested functions from same file", {
  # Create code with a nested function
  code <- "
  outer <- function() {
    inner <- function(x) { x + 1 }
    inner(5)
  }
  "
  parsed <- parse(text = code, keep.source = TRUE)
  eval(parsed[[1]])

  result <- src(outer)

  # The outer function should have a srcfile with an ID
  outer_srcfile <- result$`attr("srcref")`$`attr("srcfile")`
  expect_true(!is.null(attr(outer_srcfile, "srcfile_id")))

  # All other references should use the same ID
  id <- attr(outer_srcfile, "srcfile_id")

  # Check wholeSrcref reference
  whole_srcfile <- result$`body()`$`attr("wholeSrcref")`$`attr("srcfile")`
  expect_s3_class(whole_srcfile, "lobstr_srcfile_ref")
  expect_equal(as.character(whole_srcfile), id)
})

test_that("srcfile deduplication - reference notation displays correctly", {
  code <- parse(text = "f <- function(x) { x + 1 }", keep.source = TRUE)
  f <- eval(code[[1]])

  # Capture the output
  output <- capture.output(print(src(f)))

  # Should see the full srcfile once with @id notation
  full_srcfile_lines <- grep("<srcfilecopy> @[0-9a-f]+", output)
  expect_true(length(full_srcfile_lines) >= 1)

  # Should see reference notation (just @id without class)
  ref_lines <- grep("^[^<]*@[0-9a-f]+\\s*$", output, perl = TRUE)
  expect_true(length(ref_lines) >= 1)

  # Extract the ID from both to verify they match
  full_line <- output[full_srcfile_lines[1]]
  id_from_full <- regmatches(full_line, regexpr("@[0-9a-f]+", full_line))

  ref_line <- output[ref_lines[1]]
  id_from_ref <- regmatches(ref_line, regexpr("@[0-9a-f]+", ref_line))

  expect_equal(id_from_full, id_from_ref)
})

test_that("lobstr_srcfile_ref class has correct structure", {
  # Create a reference object directly
  ref <- new_srcfile_ref("abc123", "srcfilecopy")

  expect_s3_class(ref, "lobstr_srcfile_ref")
  expect_equal(as.character(ref), "abc123")
  expect_equal(attr(ref, "srcfile_class"), "srcfilecopy")

  # Tree label should show just @id
  label <- tree_label.lobstr_srcfile_ref(ref, list())
  expect_equal(label, "@abc123")
})

test_that("srcfile deduplication - expression with multiple elements", {
  # Parse an expression with multiple top-level elements
  code <- parse(text = c("x <- 1", "y <- 2", "z <- 3"), keep.source = TRUE)

  result <- src(code)

  # The expression should have an srcref list
  srcref_list <- result$`attr("srcref")`

  # Get the first srcfile
  first_srcfile <- srcref_list$`[[1]]`$`attr("srcfile")`
  id <- attr(first_srcfile, "srcfile_id")
  expect_type(id, "character")

  # Other elements should reference the same srcfile
  second_srcfile <- srcref_list$`[[2]]`$`attr("srcfile")`
  expect_s3_class(second_srcfile, "lobstr_srcfile_ref")
  expect_equal(as.character(second_srcfile), id)

  third_srcfile <- srcref_list$`[[3]]`$`attr("srcfile")`
  expect_s3_class(third_srcfile, "lobstr_srcfile_ref")
  expect_equal(as.character(third_srcfile), id)
})

# Deep nesting tests ----------------------------------------------------------

test_that("deep nesting - for loop with nested block", {
  code <- parse(
    text = "
  f <- function(x) {
    for (i in 1:x) {
      print(i)
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(f)

  # Should have body with nested block
  expect_true("body()" %in% names(result))

  # Should show nested block with path notation [[2]][[4]]
  # (element 2 of body is the for loop, element 4 is the body block)
  nested_block_path <- grep(
    "^\\[\\[2\\]\\]\\[\\[4\\]\\]",
    names(result$`body()`),
    value = TRUE
  )
  expect_true(length(nested_block_path) >= 1)

  # The nested block should have srcref attributes
  nested_block <- result$`body()`[[nested_block_path[1]]]
  expect_s3_class(nested_block, "lobstr_srcref")
  expect_true(
    !is.null(nested_block$`attr("srcref")`) ||
      !is.null(nested_block$`attr("wholeSrcref")`)
  )
})

test_that("deep nesting - if/else with blocks", {
  code <- parse(
    text = "
  g <- function(x) {
    if (x > 0) {
      y <- x + 1
    } else {
      y <- x - 1
    }
    y
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(g)

  # Should show both if and else blocks with path notation
  body_names <- names(result$`body()`)

  # Look for nested blocks (should be [[2]][[3]] and [[2]][[4]])
  if_block_paths <- grep(
    "^\\[\\[2\\]\\]\\[\\[3\\]\\]",
    body_names,
    value = TRUE
  )
  else_block_paths <- grep(
    "^\\[\\[2\\]\\]\\[\\[4\\]\\]",
    body_names,
    value = TRUE
  )

  expect_true(length(if_block_paths) >= 1)
  expect_true(length(else_block_paths) >= 1)

  # Both blocks should be srcref objects
  if_block <- result$`body()`[[if_block_paths[1]]]
  expect_s3_class(if_block, "lobstr_srcref")

  else_block <- result$`body()`[[else_block_paths[1]]]
  expect_s3_class(else_block, "lobstr_srcref")
})

test_that("deep nesting - nested blocks { { { } } }", {
  code <- parse(
    text = "
  h <- function() {
    {
      {
        x <- 1
      }
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(h)

  # Should have nested blocks
  expect_true("body()" %in% names(result))

  # Should have at least one nested [[2]] block
  expect_true("[[2]]" %in% names(result$`body()`))

  # That nested block should have further nesting
  nested_block <- result$`body()`$`[[2]]`
  expect_s3_class(nested_block, "lobstr_srcref")
  expect_true("[[2]]" %in% names(nested_block))
})

test_that("deep nesting - multiple top-level statements", {
  code <- parse(
    text = "
  f <- function(x) {
    a <- x + 1
    b <- a * 2
    for (i in 1:b) {
      print(i)
    }
    b
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(f)

  # Should have body with statement srcrefs
  expect_true("body()" %in% names(result))
  expect_true("attr(\"srcref\")" %in% names(result$`body()`))

  # Should show the nested for loop block
  body_names <- names(result$`body()`)
  for_block_paths <- grep(
    "\\[\\[4\\]\\]\\[\\[4\\]\\]",
    body_names,
    value = TRUE
  )
  expect_true(length(for_block_paths) >= 1)
})

test_that("deep nesting - empty function", {
  code <- parse(text = "f <- function() {}", keep.source = TRUE)

  eval(code[[1]])
  result <- src(f)

  # Should still have structure
  expect_type(result, "list")
  expect_s3_class(result, "lobstr_srcref")
  expect_true("attr(\"srcref\")" %in% names(result))
})

test_that("deep nesting - very deep structure respects max_depth", {
  # Create deeply nested structure
  code <- parse(
    text = "
  f <- function(x) {
    for (i in 1:x) {
      if (i > 0) {
        while (i < 10) {
          for (j in 1:i) {
            print(j)
          }
          break
        }
      }
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])

  # With low max_depth, should truncate
  result_shallow <- src(f, max_depth = 2)
  expect_type(result_shallow, "list")

  # With high max_depth, should show more nesting
  result_deep <- src(f, max_depth = 10)
  expect_type(result_deep, "list")

  # Deep version should have more nested paths
  shallow_paths <- names(result_shallow)
  deep_paths <- names(result_deep)

  # Both should be valid results
  expect_true(length(shallow_paths) > 0)
  expect_true(length(deep_paths) > 0)
})

test_that("deep nesting - intermediate calls without srcrefs are omitted", {
  code <- parse(
    text = "
  f <- function(x) {
    for (i in 1:x) {
      print(i)
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])

  # Capture output
  output <- capture.output(print(src(f)))

  # Should NOT see intermediate <language> nodes
  language_nodes <- grep("<language>", output, value = TRUE)
  expect_equal(length(language_nodes), 0)

  # Should see collapsed path notation
  collapsed_paths <- grep(
    "\\[\\[\\d+\\]\\]\\[\\[\\d+\\]\\]",
    output,
    value = TRUE
  )
  expect_true(length(collapsed_paths) > 0)
})

test_that("deep nesting - while loop with nested block", {
  code <- parse(
    text = "
  f <- function(x) {
    i <- 0
    while (i < x) {
      print(i)
      i <- i + 1
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(f)

  # Should have nested block for while body
  body_names <- names(result$`body()`)
  while_block_paths <- grep(
    "\\[\\[3\\]\\]\\[\\[3\\]\\]",
    body_names,
    value = TRUE
  )
  expect_true(length(while_block_paths) >= 1)
})

test_that("deep nesting - repeat loop with nested block", {
  code <- parse(
    text = "
  f <- function(x) {
    repeat {
      print(x)
      break
    }
  }
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(f)

  # Should have nested block for repeat body
  expect_true("body()" %in% names(result))
  body_names <- names(result$`body()`)

  # Look for nested block path
  nested_paths <- grep("^\\[\\[2\\]\\]\\[\\[2\\]\\]", body_names, value = TRUE)
  expect_true(length(nested_paths) >= 1)
})

test_that("deep nesting - switch statement with blocks", {
  code <- parse(
    text = '
  f <- function(x) {
    switch(x,
      a = {
        print("a")
      },
      b = {
        print("b")
      }
    )
  }
  ',
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(f)

  # Should have structure
  expect_type(result, "list")
  expect_s3_class(result, "lobstr_srcref")
})

test_that("deep nesting - quoted expressions with nested blocks", {
  code <- parse(
    text = "
  x <- quote({
    for (i in 1:3) {
      print(i)
    }
  })
  ",
    keep.source = TRUE
  )

  eval(code[[1]])
  result <- src(x)

  # Should show nested structure
  expect_type(result, "list")

  # Should have nested block paths
  if (!is.null(result) && length(result) > 0) {
    all_names <- names(unlist(result, recursive = TRUE))
    nested_paths <- grep(
      "\\[\\[\\d+\\]\\]\\[\\[\\d+\\]\\]",
      all_names,
      value = TRUE
    )
    # Might or might not have nested paths depending on how quote() preserves srcrefs
    expect_true(length(nested_paths) >= 0)
  }
})

#' Display the abstract syntax tree
#'
#' This is a useful alternative to `str()` for expression objects.
#'
#' @param x An expression to display. Input is automatically quoted,
#'   use `!!` to unquote if you have already captured an expression object.
#' @family object inspectors
#' @export
#' @examples
#' # Leaves
#' ast(1)
#' ast(x)
#'
#' # Simple calls
#' ast(f())
#' ast(f(x, 1, g(), h(i())))
#' ast(f()())
#' ast(f(x)(y))
#'
#' ast((x + 1))
#'
#' # Displaying expression already stored in object
#' x <- quote(a + b + c)
#' ast(x)
#' ast(!!x)
#'
#' # All operations have this same structure
#' ast(if (TRUE) 3 else 4)
#' ast(y <- x * 10)
#' ast(function(x = 1, y = 2) { x + y } )
#'
#' # Operator precedence
#' ast(1 * 2 + 3)
#' ast(!1 + !1)
ast <- function(x) {
  expr <- enexpr(x)
  new_raw(ast_tree(expr))
}

ast_tree <- function(x, layout = box_chars()) {
  if (is_quosure(x)) {
    x <- quo_squash(x)
  }

  # base cases
  if (rlang::is_syntactic_literal(x)) {
    return(ast_leaf_constant(x))
  } else if (is_symbol(x)) {
    return(ast_leaf_symbol(x))
  } else if (!is.pairlist(x) && !is.call(x)) {
    return(paste0("<inline ", paste0(class(x), collapse = "/"), ">"))
  }

  # recursive case
  subtrees <- lapply(x, ast_tree, layout = layout)
  subtrees <- name_subtree(subtrees)

  n <- length(x)
  if (n == 0) {
    character()
  } else if (n == 1) {
    str_indent(subtrees[[1]], paste0(layout$n, layout$h), "  ")
  } else {
    c(
      str_indent(
        subtrees[[1]],
        paste0(layout$n, layout$h),
        paste0(layout$v, " ")
      ),
      unlist(lapply(
        subtrees[-c(1, n)],
        str_indent,
        paste0(layout$j, layout$h),
        paste0(layout$v, " ")
      )),
      str_indent(subtrees[[n]], paste0(layout$l, layout$h), "  ")
    )
  }
}

name_subtree <- function(x) {
  nm <- names(x)
  if (is.null(nm)) {
    return(x)
  }

  has_name <- nm != ""
  label <- paste0(crayon::italic(grey(nm)), " = ")
  indent <- str_dup(" ", nchar(nm) + 3)

  x[has_name] <- Map(str_indent, x[has_name], label[has_name], indent[has_name])
  x
}

ast_leaf_symbol <- function(x) {
  x <- as.character(x)
  if (!is.syntactic(x)) {
    x <- encodeString(x, quote = "`")
  }

  crayon::bold(crayon::magenta(x))
}
ast_leaf_constant <- function(x) {
  if (is.complex(x)) {
    paste0(Im(x), "i")
  } else {
    deparse(x)
  }
}

is.syntactic <- function(x) make.names(x) == x

test_that("quosures print same as expressions", {
  expect_equal(ast_tree(quo(x)), ast_tree(expr(x)))
})

test_that("can print complex expression", {
  skip_on_os("windows")

  x <- expr(function(x) if (x > 1) f(y$x, "x", g()))
  expect_snapshot({
    ast(!!x)
  })
})

test_that("can print complex expression without unicode", {
  old <- options(lobstr.fancy.tree = FALSE)
  on.exit(options(old))

  x <- expr(function(x) if (x > 1) f(y$x, "x", g()))
  expect_snapshot({
    ast(!!x)
  })
})

test_that("can print scalar expressions nicely", {
  old <- options(lobstr.fancy.tree = FALSE)
  on.exit(options(old))

  x <- expr(list(
    logical = c(FALSE, TRUE, NA),
    integer = 1L,
    double = 1,
    character = "a",
    complex = 1i
  ))
  expect_snapshot({
    ast(!!x)
  })
})

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}


# CLI ---------------------------------------------------------------------

box_chars <- function() {
  fancy <- getOption("lobstr.fancy.tree") %||% l10n_info()$`UTF-8`
  orange <- crayon::make_style("orange")

  if (fancy) {
    list(
      "h" = "\u2500", # - horizontal
      "hd" = "\u2504", # - horizontal dotted
      "v" = "\u2502", # | vertical
      "vd" = "\u250A", # | vertical dotted
      "l" = "\u2514", # \ leaf
      "j" = "\u251C", # + junction
      "n" = orange("\u2588") # X node
    )
  } else {
    list(
      "h" = "-",
      "hd" = "-", # Just use normal chars for dotted
      "v" = "|",
      "vd" = "|",
      "l" = "\\",
      "j" = "+",
      "n" = orange("o")
    )
  }
}

grey <- function(...) {
  crayon::make_style(grDevices::grey(0.5), grey = TRUE)(...)
}

# string -----------------------------------------------------------------

str_dup <- function(x, n) {
  vapply(n, function(i) paste0(rep(x, i), collapse = ""), character(1))
}

str_indent <- function(x, first, rest) {
  if (length(x) == 0) {
    character()
  } else if (length(x) == 1) {
    paste0(first, x)
  } else {
    c(
      paste0(first, x[[1]]),
      paste0(rest, x[-1L])
    )
  }
}

str_truncate <- function(x, n) {
  too_long <- nchar(x, type = "width") > n
  x[too_long] <- paste0(substr(x[too_long], 1, n - 3), "...")
  x
}

new_raw <- function(x) {
  structure(x, class = "lobstr_raw")
}

#' @export
print.lobstr_raw <- function(x, ...) {
  cat(paste(x, "\n", collapse = ""), sep = "")
  invisible(x)
}

cat_line <- function(...) {
  cat(paste0(..., "\n", collapse = ""))
}

# Snapshot transformer to scrub src() output for deterministic snapshots
scrub_src_transform <- function(lines) {
  lines <- gsub('filename: "[^"]+"', 'filename: "<scrubbed>"', lines)
  lines <- gsub('directory: "[^"]+"', 'directory: "<scrubbed>"', lines)
  lines <- gsub('timestamp: "[^"]+"', 'timestamp: "<scrubbed>"', lines)
  lines <- gsub('wd: "[^"]+"', 'wd: "<scrubbed>"', lines)
  lines
}

with_srcref <- function(code, env = parent.frame(), file = NULL) {
  if (is.null(file)) {
    file <- tempfile("test_srcref", fileext = ".R")
    on.exit(unlink(file), add = TRUE)
  }

  writeLines(code, file)
  source(file, local = env, keep.source = TRUE)
}

parse_with_srcref <- function(code) {
  parse(text = code, keep.source = TRUE)
}

simple_function_with_srcref <- function() {
  code <- c(
    "test_func <- function(x, y) {",
    "  x + y",
    "}"
  )

  env <- new.env(parent = baseenv())
  with_srcref(code, env = env)
  env$test_func
}

multi_statement_function_with_srcref <- function() {
  code <- c(
    "multi_func <- function(x) {",
    "  a <- x + 1",
    "  b <- a * 2",
    "  c <- b - 3",
    "  c",
    "}"
  )

  env <- new.env(parent = baseenv())
  with_srcref(code, env = env)
  env$multi_func
}

#' Inspect an object
#'
#' `sxp(x)` is similar to `.Internal(inspect(x))`, recursing into the C data
#' structures underlying any R object. The main difference is the output is a
#' little more compact, it recurses fully, and avoids getting stuck in infinite
#' loops by using a depth-first search. It also returns a list that you can
#' compute with, and carefully uses colour to highlight the most important
#' details.
#'
#' The name `sxp` comes from `SEXP`, the name of the C data structure that
#' underlies all R objects.
#'
#' @param x Object to inspect
#' @param max_depth Maximum depth to recurse. Use `max_depth = Inf` (with care!)
#'   to recurse as deeply as possible. Skipped elements will be shown as `...`.`
#' @param expand Optionally, expand components of the true that are usually
#'   suppressed. Use:
#'
#'   * "character" to show underlying entries in the global string pool.
#'   * "environment" to show binding components without any side effects (e.g. promises or active bindings).
#'   * "altrep" to show the underlying data.
#'   * "call" to show the full AST (but [ast()] is usually superior)
#'   * "bytecode" to show generated bytecode.
#' @family object inspectors
#' @export
#' @examples
#' x <- list(
#'   TRUE,
#'   1L,
#'   runif(100),
#'   "3"
#' )
#' sxp(x)
#'
#' # Expand "character" to see underlying CHARSXP entries in the global
#' # string pool
#' x <- c("banana", "banana", "apple", "banana")
#' sxp(x)
#' sxp(x, expand = "character")
#'
#' # Expand altrep to see underlying data
#' x <- 1:10
#' sxp(x)
#' sxp(x, expand = "altrep")
#'
#' # Expand environments to see promise expressions without forcing
#' e <- new.env(parent = emptyenv())
#' delayedAssign("x", 1 + 1, assign.env = e)
#'
#' sxp(e)
#' sxp(e, expand = "environment")

sxp <- function(x, expand = character(), max_depth = 5L) {
  opts <- c("character", "altrep", "environment", "call", "bytecode")
  if (any(!expand %in% opts)) {
    abort(
      sprintf(
        "`expand` must contain only values from: '%s'.",
        paste(opts, collapse = "', '")
      )
    )
  }

  obj_inspect_(
    x,
    max_depth - 1L,
    opts[[1]] %in% expand,
    opts[[2]] %in% expand,
    opts[[3]] %in% expand,
    opts[[4]] %in% expand,
    opts[[5]] %in% expand
  )
}

#' @export
format.lobstr_inspector <- function(x, ..., depth = 0, name = NA) {
  indent <- paste0(rep("  ", depth), collapse = "")

  id <- crayon::bold(attr(x, "id"))
  if (!is_testing() && !is_placeholder(x)) {
    addr <- paste0(":", crayon::silver(attr(x, "addr")))
  } else {
    addr <- ""
  }

  type <- attr(x, "type")

  if (attr(x, "has_seen")) {
    desc <- paste0("[", attr(x, "id"), addr, "]")
  } else {
    if (sexp_is_vector(type)) {
      length <- paste0("[", attr(x, "length"), "]")
    } else {
      length <- NULL
    }

    if (!is.null(attr(x, "value"))) {
      value <- paste0(": ", attr(x, "value"))
    } else {
      value <- NULL
    }

    if (!is_testing() && !is_placeholder(x)) {
      no_references <- attr(x, "no_references")
      maybe_shared <- attr(x, "maybe_shared")
      if (no_references == 1) {
        references <- "refs:0"
      } else if (maybe_shared == 0) {
        references <- "refs:1"
      } else {
        references <- "refs:2+"
      }
    } else {
      references <- NULL
    }

    # show altrep, object, named etc
    sxpinfo <- paste0(
      if (attr(x, "altrep")) "altrep ",
      if (attr(x, "object")) "object ",
      references
    )

    # Placeholders don't show sxpinfo or id (they can't be referenced)
    if (is_placeholder(x)) {
      desc <- paste0(
        "<",
        crayon::cyan(type),
        value,
        ">"
      )
    } else {
      desc <- paste0(
        "[",
        id,
        addr,
        "] ",
        "<",
        crayon::cyan(type),
        length,
        value,
        "> ",
        "(",
        sxpinfo,
        ")"
      )
    }
  }

  name <- if (!identical(name, "")) {
    paste0(crayon::italic(crayon::silver(name)), " ")
  }

  paste0(indent, name, desc)
}

#' @export
print.lobstr_inspector <- function(x, ..., depth = 0, name = "") {
  cat_line(format(x, depth = depth, name = name))

  if (isTRUE(attr(x, "skip"))) {
    indent <- paste0(rep("  ", depth + 1), collapse = "")
    cat_line(indent, crayon::silver("..."))
  }

  for (i in seq_along(x)) {
    print(x[[i]], depth = depth + 1, name = names(x)[[i]])
  }
}

sxp_view <- function(x, expand = character()) {
  if (!"tools:rstudio" %in% search()) {
    abort("Can only be called from within RStudio")
  }

  env <- as.environment("tools:rstudio")

  old_opt <- options(crayon.enabled = FALSE)
  on.exit(options(old_opt), add = TRUE)

  old_fun <- env$.rs.explorer.objectDesc
  on.exit(env$.rs.addFunction("explorer.objectDesc", old_fun), add = TRUE)

  assign(".rs.explorer.objectDesc", envir = env, function(x) {
    if (inherits(x, "lobstr_inspector")) {
      format.lobstr_inspector(x)
    } else {
      old_fun(x)
    }
  })

  obj <- sxp(x, expand = expand)
  env$.rs.viewHook(NULL, obj, "Object inspector")

  # explorer.objectDesc() is called lazily so this is a crude hack
  Sys.sleep(10)
}

# helpers -----------------------------------------------------------------

sexp_is_vector <- function(x) {
  x %in%
    c(
      "LGLSXP",
      "INTSXP",
      "REALSXP",
      "STRSXP",
      "RAWSXP",
      "CPLXSXP",
      "VECSXP",
      "EXPRSXP"
    )
}

# Placeholder nodes do not have any inspectable properties such as refcount or
# address
is_placeholder <- function(x) {
  !nzchar(attr(x, "addr")) || identical(attr(x, "type"), "NILSXP")
}

test_that("computes spanning tree", {
  x <- 1:10
  y <- list(x, x, x)
  obj <- sxp(y)

  expect_false(attr(obj[[1]], "has_seen"))
  expect_true(attr(obj[[2]], "has_seen"))
})

test_that("captures names of special environments", {
  x <- list(
    emptyenv(),
    baseenv(),
    globalenv()
  )
  obj <- sxp(x)
  expect_equal(attr(obj[[1]], "value"), "empty")
  expect_equal(attr(obj[[2]], "value"), "base")
  expect_equal(attr(obj[[3]], "value"), "global")
})

test_that("captures names of lists", {
  x <- list(a = 1, b = 2, c = 3)
  obj <- sxp(x)
  expect_named(obj, c(names(x), "_attrib"))
})

test_that("can expand lists", {
  x <- c("xxx", "xxx", "y")
  obj <- sxp(x, expand = "character")

  expect_length(obj, 3)
  expect_equal(attr(obj[[1]], "ref"), attr(obj[[2]], "ref"))
})

test_that("can inspect active bindings", {
  e <- new.env(hash = FALSE)
  env_bind_active(e, f = function() stop("!"))

  x <- sxp(e)
  expect_named(x, c("f", "_enclos"))
})

test_that("snapshots environment binding types", {
  e <- new.env(parent = emptyenv(), hash = FALSE)

  # value
  e$value <- 1

  # missing argument binding
  env_bind(e, missing = missing_arg())

  # delayed and forced promise bindings
  delayedAssign("delayed", 1 + 1, assign.env = e)
  delayedAssign("forced", 1 + 1, assign.env = e)
  invisible(e$forced)

  # active binding
  env_bind_active(e, active = function() 42)

  expect_snapshot({
    print(sxp(e))
    print(sxp(e, expand = "environment", max_depth = 6L))
  })
})

# Regression tests --------------------------------------------------------

test_that("can inspect all atomic vectors", {
  x <- list(
    TRUE,
    1L,
    1,
    "3",
    1i,
    raw(1)
  )
  expect_snapshot(sxp(x))
})

test_that("can inspect functions", {
  f <- function(x, y = 1, ...) x + 1
  attr(f, "srcref") <- NULL
  environment(f) <- globalenv()

  expect_snapshot(sxp(f))
})

test_that("can inspect environments", {
  e1 <- new.env(parent = emptyenv(), size = 5L)
  e1$x <- 10
  e1$y <- e1

  e2 <- new.env(parent = e1, size = 5L)

  expect_snapshot({
    print(sxp(e2))
    print(sxp(e2, expand = "environment", max_depth = 5L))
  })
})

test_that("can expand altrep", {
  skip_if_not(getRversion() >= "3.5")
  skip_if_not(.Machine$sizeof.pointer == 8) # _class RAWSXP has different size

  expect_snapshot({
    x <- 1:10
    print(sxp(x, expand = "altrep", max_depth = 4L))
  })
})

test_that("can inspect cons cells", {
  expect_snapshot({
    cell <- new_node(1, 2)
    sxp(cell)

    non_nil_terminated_list <- new_node(1, new_node(2, 3))
    sxp(non_nil_terminated_list)
  })
})

test_that("fix error message when `expand` argument contains invalid classes", {
  expect_snapshot(error = TRUE, {
    sxp(1, expand = "invalid_class")
  })
})

#' Display tree of references
#'
#' This tree display focusses on the distinction between names and values.
#' For each reference-type object (lists, environments, and optional character
#' vectors), it displays the location of each component. The display
#' shows the connection between shared references using a locally unique id.
#'
#' @param ... One or more objects
#' @param character If `TRUE`, show references from character vector in to
#'   global string pool
#' @export
#' @family object inspectors
#' @examples
#' x <- 1:100
#' ref(x)
#'
#' y <- list(x, x, x)
#' ref(y)
#' ref(x, y)
#'
#' e <- new.env()
#' e$e <- e
#' e$x <- x
#' e$y <- list(x, e)
#' ref(e)
#'
#' # Can also show references to global string pool if requested
#' ref(c("x", "x", "y"))
#' ref(c("x", "x", "y"), character = TRUE)
ref <- function(..., character = FALSE) {
  x <- list(...)
  seen <- child_env(emptyenv(), `__next_id` = 1)

  out <- lapply(x, ref_tree, character = character, seen = seen)

  n <- length(x)
  if (n > 1) {
    out[-n] <- lapply(out[-n], function(x) c(x, ""))
  }
  new_raw(unlist(out))
}

ref_tree <- function(
  x,
  character = FALSE,
  seen = child_env(emptyenv()),
  layout = box_chars()
) {
  addr <- obj_addr(x)
  has_seen <- env_has(seen, addr)
  id <- obj_id(seen, addr)

  desc <- obj_desc(addr, type_sum(x), has_seen, id)

  # Not recursive or already seen
  if (!has_references(x, character) || has_seen) {
    return(desc)
  }

  # Remove classes to avoid custom methods (note that environments cannot be unclasse()ed)
  attr(x, "class") <- NULL

  # recursive cases
  if (is.list(x)) {
    subtrees <- lapply(
      x,
      ref_tree,
      layout = layout,
      seen = seen,
      character = character
    )
  } else if (is.environment(x)) {
    subtrees <- lapply(
      as.list(x, all.names = TRUE),
      ref_tree,
      layout = layout,
      seen = seen,
      character = character
    )
  } else if (is.character(x)) {
    subtrees <- ref_tree_chr(x, layout = layout, seen = seen)
  }
  subtrees <- name_subtree(subtrees)

  self <- str_indent(desc, paste0(layout$n, " "), paste0(layout$v, " "))

  n <- length(subtrees)
  if (n == 0) {
    return(self)
  }
  c(
    self,
    unlist(lapply(
      subtrees[-n],
      str_indent,
      paste0(layout$j, layout$h),
      paste0(layout$v, " ")
    )),
    str_indent(subtrees[[n]], paste0(layout$l, layout$h), "  ")
  )
}

type_sum <- function(x) {
  if (is_installed("pillar")) {
    pillar::type_sum(x)
  } else {
    typeof(x)
  }
}

obj_desc <- function(addr, type, has_seen, id) {
  if (has_seen) {
    paste0("[", grey(paste0(id, ":", addr)), "]")
  } else {
    paste0("[", crayon::bold(id), ":", addr, "] ", "<", type, ">")
  }
}

has_references <- function(x, character = FALSE) {
  is_list(x) || is.environment(x) || (character && is_character(x))
}

ref_tree_chr <- function(
  x,
  layout = box_chars(),
  seen = child_env(emptyenv())
) {
  addrs <- obj_addrs(x)

  has_seen <- logical(length(x))
  ids <- integer(length(x))
  for (i in seq_along(addrs)) {
    has_seen[[i]] <- env_has(seen, addrs[[i]])
    ids[[i]] <- obj_id(seen, addrs[[i]])
  }

  type <- paste0('string: "', str_truncate(x, 10), '"')

  out <- Map(obj_desc, addrs, type, has_seen, ids)
  names(out) <- names(x)
  out
}

obj_id <- function(env, ref) {
  if (env_has(env, ref)) {
    env_get(env, ref)
  } else {
    id <- env_get(env, "__next_id")
    env_poke(env, "__next_id", id + 1)
    env_poke(env, ref, id)

    id
  }
}

#' Calculate the size of an object.
#'
#' `obj_size()` computes the size of an object or set of objects;
#' `obj_sizes()` breaks down the individual contribution of multiple objects
#' to the total size.
#'
#' @section Compared to `object.size()`:
#' Compared to [object.size()], `obj_size()`:
#'
#' * Accounts for all types of shared values, not just strings in
#'   the global string pool.
#'
#' * Includes the size of environments (up to `env`)
#'
#' * Accurately measures the size of ALTREP objects.
#'
#' @section Environments:
#' `obj_size()` attempts to take into account the size of the
#' environments associated with an object. This is particularly important
#' for closures and formulas, since otherwise you may not realise that you've
#' accidentally captured a large object. However, it's easy to over count:
#' you don't want to include the size of every object in every environment
#' leading back to the [emptyenv()]. `obj_size()` takes
#' a heuristic approach: it never counts the size of the global environment,
#' the base environment, the empty environment, or any namespace.
#'
#' Additionally, the `env` argument allows you to specify another
#' environment at which to stop. This defaults to the environment from which
#' `obj_size()` is called to prevent double-counting of objects created
#' elsewhere.
#'
#' @export
#' @param ... Set of objects to compute size.
#' @param env Environment in which to terminate search. This defaults to the
#'   current environment so that you don't include the size of objects that
#'   are already stored elsewhere.
#'
#'   Regardless of the value here, `obj_size()` never looks past the
#'   global or base environments.
#'
#' @return An estimate of the size of the object, in bytes.
#' @examples
#' # obj_size correctly accounts for shared references
#' x <- runif(1e4)
#' obj_size(x)
#'
#' z <- list(a = x, b = x, c = x)
#' obj_size(z)
#'
#' # this means that object size is not transitive
#' obj_size(x)
#' obj_size(z)
#' obj_size(x, z)
#'
#' # use obj_size() to see the unique contribution of each component
#' obj_sizes(x, z)
#' obj_sizes(z, x)
#' obj_sizes(!!!z)
#'
#' # obj_size() also includes the size of environments
#' f <- function() {
#'   x <- 1:1e4
#'   a ~ b
#' }
#' obj_size(f())
#'
#' #' # In R 3.5 and greater, `:` creates a special "ALTREP" object that only
#' # stores the first and last elements. This will make some vectors much
#' # smaller than you'd otherwise expect
#' obj_size(1:1e6)
obj_size <- function(..., env = parent.frame()) {
  dots <- list2(...)
  size <- obj_size_(dots, env, size_node(), size_vector())
  new_bytes(size)
}

#' @rdname obj_size
#' @export
obj_sizes <- function(..., env = parent.frame()) {
  dots <- list2(...)
  size <- obj_csize_(dots, env, size_node(), size_vector())
  names(size) <- names(dots)
  new_bytes(size)
}

size_node <- function(x) as.vector(utils::object.size(quote(expr = )))
size_vector <- function(x) as.vector(utils::object.size(logical()))

new_bytes <- function(x) {
  structure(x, class = "lobstr_bytes")
}

#' @export
format.lobstr_bytes <- function(x, ...) {
  prettyunits::pretty_bytes(unclass(x))
}

#' @export
print.lobstr_bytes <- function(x, ...) {
  fx <- format(x)

  if (length(x) == 1) {
    cat_line(fx)
  } else {
    if (!is.null(names(x))) {
      cat_line(format(names(x)), ": ", fx)
    } else {
      cat_line("* ", fx)
    }
  }

  invisible(x)
}

#' @export
c.lobstr_bytes <- function(...) {
  new_bytes(NextMethod())
}

#' @export
`[.lobstr_bytes` <- function(...) {
  new_bytes(NextMethod())
}

# Helpers for interactive exploration -------------------------------------

comp <- function(x) {
  base <- utils::object.size(x)
  lobstr <- obj_size(x)
  c(base = base, lobstr = lobstr, diff = base - lobstr)
}
insp <- function(x) {
  eval(quote(.Internal(inspect(x))))
}

expect_same <- function(x, ...) {
  lab <- as.character(expr_text(enexpr(x)))
  act <- as.vector(obj_size(x))
  exp <- as.vector(object.size(x))

  msg <- sprintf(
    "`obj_size(%s)` is %s, not %s (\u0394%+i)",
    lab,
    act,
    exp,
    act - exp
  )
  expect(identical(act, exp), msg)
  invisible(act)
}

# S3 methods --------------------------------------------------------------

test_that("combined bytes are aligned", {
  expect_snapshot({
    new_bytes(c(400, 400000))
  })
})

# Compatibility with base ---------------------------------------------------

test_that("size correct for length one vectors", {
  expect_same(1)
  expect_same(1L)
  expect_same("abc")
  expect_same(paste(rep("banana", 100), collapse = ""))
  expect_same(charToRaw("a"))
  expect_same(5 + 1i)
})

test_that("size scales correctly with length (accounting for vector pool)", {
  expect_same(numeric())
  expect_same(1)
  expect_same(2)
  expect_same(c(1:10))
  expect_same(c(1:1000))
})

test_that("size of list computed recursively", {
  expect_same(list())
  expect_same(as.list(1))
  expect_same(as.list(1:2))
  expect_same(as.list(1:3))

  expect_same(list(list(list(list(list())))))
})

test_that("size of symbols same as base", {
  expect_same(quote(x))
  expect_same(quote(asfsadfasdfasdfds))
})

test_that("size of pairlists same as base", {
  expect_same(pairlist())
  expect_same(pairlist(1))
  expect_same(pairlist(1, 2))
  expect_same(pairlist(1, 2, 3))
  expect_same(pairlist(1, 2, 3, 4))
})

test_that("don't crash with large pairlists", {
  n <- 1e5
  x <- pairlist(1)
  xn <- as.pairlist(rep(1, n))
  expect_equal(obj_size(xn), n * obj_size(x))
})

test_that("size of S4 objects same as base", {
  Z <- methods::setClass("Z", slots = c(x = "integer"))
  z <- Z(x = 1L)

  expect_same(z)
})

test_that("size of attributes included in object size", {
  expect_same(c(x = 1))
  expect_same(list(x = 1))
  expect_same(c(x = "y"))
})

test_that("duplicated CHARSXPS only counted once", {
  expect_same("x")
  expect_same(c("x", "y", "x"))
  expect_same(c("banana", "banana", "banana"))
})

test_that("obj_sizes computes relative size", {
  x <- 1:10 + 1
  out <- obj_sizes(x, x)
  expect_equal(out[1], obj_size(x))
  expect_equal(out[2], new_bytes(0))
})

# Improved behaviour for shared components ------------------------------------
test_that("shared components only counted once", {
  x <- 1:1e3
  z <- list(x, x, x)

  expect_equal(obj_size(z), obj_size(x) + obj_size(vector("list", 3)))
})

test_that("size of closures same as base", {
  f <- function() NULL
  attributes(f) <- NULL # zap srcrefs
  environment(f) <- emptyenv()
  expect_same(f)
})

# Improved behaviour for ALTREP objects -----------------------------------

test_that("altrep size measured correctly", {
  skip_if_not(getRversion() > "3.5.0")

  # Currently reported size is 640 B
  # If regular vector would be 4,000,040 B
  # This test is conservative so shouldn't fail in case representation
  # changes in the future
  expect_true(obj_size(1:1e6) < 10000)
})

test_that("can compute size of deferred string vectors", {
  x <- 1:10
  names(x) <- 10:1
  y <- names(x)
  obj_size(y)

  # Just assert that it doesn't crash
  succeed("Didn't crash")
})

# Environment sizes -----------------------------------------------------------
test_that("terminal environments have size zero", {
  expect_equal(obj_size(globalenv()), new_bytes(0))
  expect_equal(obj_size(baseenv()), new_bytes(0))
  expect_equal(obj_size(emptyenv()), new_bytes(0))

  expect_equal(obj_size(asNamespace("stats")), new_bytes(0))
})

test_that("environment size computed recursively", {
  e <- new.env(parent = emptyenv())
  e_size <- obj_size(e)

  f <- new.env(parent = e)
  obj_size(f)
  expect_equal(obj_size(f), 2 * obj_size(e))
})

test_that("size of function includes environment", {
  f <- function() {
    y <- 1:1e3
    a ~ b
  }
  g <- function() {
    y <- 1:1e3
    function() 10
  }

  expect_true(obj_size(f()) > obj_size(1:1e3))
  expect_true(obj_size(g()) > obj_size(1:1e3))
})

test_that("size doesn't include parents of current environment", {
  x <- c(1:1e4)
  embedded <- (function() {
    g <- function() {
      x <- c(1:1e3)
      a ~ b
    }
    obj_size(g())
  })()

  expect_true(embedded < obj_size(x))
})

test_that("support dots in closure environments", {
  fn <- (function(...) function() NULL)(foo)
  expect_error(obj_size(fn), NA)
})

test_that("supports cons cells", {
  cell <- new_node(1, 2)
  expect_equal(
    obj_size(cell),
    obj_size(new_node(NULL, NULL)) + obj_size(1) + obj_size(2)
  )

  non_nil_terminated_list <- new_node(1, new_node(2, 3))
  expect_equal(
    obj_size(non_nil_terminated_list),
    obj_size(new_node(1, NULL)) + obj_size(cell)
  )
})

test_that("can size environment with loop binding (#48)", {
  fn <- function() {
    for (i in 1) {
      lobstr::obj_size(environment(), env = globalenv())
    }
  }

  # Compiling causes `i` to be an immediate binding
  fn <- compiler::cmpfun(fn)

  # Loop bindings use internal representation that previously caused
  # "bad binding access" errors when using CAR/TAG accessors
  expect_no_error(fn())
})

# Generated by cpp11: do not edit by hand

obj_addr_ <- function(name, env) {
  .Call(`_lobstr_obj_addr_`, name, env)
}

obj_addrs_ <- function(x) {
  .Call(`_lobstr_obj_addrs_`, x)
}

obj_inspect_ <- function(x, max_depth, expand_char, expand_altrep, expand_env, expand_call, expand_bytecode) {
  .Call(`_lobstr_obj_inspect_`, x, max_depth, expand_char, expand_altrep, expand_env, expand_call, expand_bytecode)
}

init_library <- function(env) {
  invisible(.Call(`_lobstr_init_library`, env))
}

v_size <- function(n, element_size) {
  .Call(`_lobstr_v_size`, n, element_size)
}

obj_size_ <- function(objects, base_env, sizeof_node, sizeof_vector) {
  .Call(`_lobstr_obj_size_`, objects, base_env, sizeof_node, sizeof_vector)
}

obj_csize_ <- function(objects, base_env, sizeof_node, sizeof_vector) {
  .Call(`_lobstr_obj_csize_`, objects, base_env, sizeof_node, sizeof_vector)
}

# src() shows closure with srcref and wholeSrcref

    Code
      f <- simple_function_with_srcref()
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:14 - 3:1
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines<chr [3]>: "test_func <-...", "  x + y", "}"
      │   ├─parseData<int [160]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:29 - 1:29
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 2:3 - 2:7
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 3:1
          └─attr("srcfile"): @001

# src() shows multi-statement function

    Code
      f <- multi_statement_function_with_srcref()
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:15 - 6:1
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines<chr [6]>: "multi_func <...", "  a <- x + 1", "  b <- a * 2", ...
      │   ├─parseData<int [352]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:27 - 1:27
        │ │ └─attr("srcfile"): @001
        │ ├─[[2]]: <srcref>
        │ │ ├─location: 2:3 - 2:12
        │ │ └─attr("srcfile"): @001
        │ ├─[[3]]: <srcref>
        │ │ ├─location: 3:3 - 3:12
        │ │ └─attr("srcfile"): @001
        │ ├─[[4]]: <srcref>
        │ │ ├─location: 4:3 - 4:12
        │ │ └─attr("srcfile"): @001
        │ └─[[5]]: <srcref>
        │   ├─location: 5:3 - 5:3
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 6:1
          └─attr("srcfile"): @001

# src() shows quoted function with nested body

    Code
      with_srcref("x <- quote(function() {})")
      src(x)
    Output
      <quoted_function>
      ├─[[3]]: <{>
      │ ├─attr("srcref"): <list>
      │ │ └─[[1]]: <srcref>
      │ │   ├─location: 1:23 - 1:23
      │ │   └─attr("srcfile"): <srcfilecopy> @001
      │ │     ├─Enc: "unknown"
      │ │     ├─filename: "<scrubbed>"
      │ │     ├─fixedNewlines: TRUE
      │ │     ├─isFile: TRUE
      │ │     ├─lines: "x <- quote(function() {})"
      │ │     ├─parseData<int [128]>: 1, 1, 1, ......
      │ │     ├─timestamp: "<scrubbed>"
      │ │     └─wd: "<scrubbed>"
      │ ├─attr("srcfile"): @001
      │ └─attr("wholeSrcref"): <srcref>
      │   ├─location: 1:0 - 1:24
      │   └─attr("srcfile"): @001
      └─[[4]]: <srcref>
        ├─location: 1:12 - 1:24
        └─attr("srcfile"): @001

# src() shows quoted function body directly

    Code
      with_srcref("x <- quote(function() {})")
      src(x[[3]])
    Output
      <{>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:23 - 1:23
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: TRUE
      │     ├─lines: "x <- quote(function() {})"
      │     ├─parseData<int [128]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 1:24
        └─attr("srcfile"): @001

# src() shows quoted function with arguments

    Code
      with_srcref("x <- quote(function(a, b) {})")
      src(x)
    Output
      <quoted_function>
      ├─[[3]]: <{>
      │ ├─attr("srcref"): <list>
      │ │ └─[[1]]: <srcref>
      │ │   ├─location: 1:27 - 1:27
      │ │   └─attr("srcfile"): <srcfilecopy> @001
      │ │     ├─Enc: "unknown"
      │ │     ├─filename: "<scrubbed>"
      │ │     ├─fixedNewlines: TRUE
      │ │     ├─isFile: TRUE
      │ │     ├─lines: "x <- quote(function(a, b) {})"
      │ │     ├─parseData<int [152]>: 1, 1, 1, ......
      │ │     ├─timestamp: "<scrubbed>"
      │ │     └─wd: "<scrubbed>"
      │ ├─attr("srcfile"): @001
      │ └─attr("wholeSrcref"): <srcref>
      │   ├─location: 1:0 - 1:28
      │   └─attr("srcfile"): @001
      └─[[4]]: <srcref>
        ├─location: 1:12 - 1:28
        └─attr("srcfile"): @001

# src() shows srcref with parsed field when positions differ

    Code
      srcfile <- srcfilecopy("test.R", c("x <- function() {",
        "  # A long comment that spans", "  # multiple lines", "  y <- 1", "}"))
      synthetic_srcref <- structure(c(2L, 3L, 4L, 8L, 3L, 8L, 1L, 5L), class = "srcref",
      srcfile = srcfile)
      src(synthetic_srcref)
    Output
      <srcref>
      ├─location: 2:3 - 4:8
      ├─parsed: 1:3 - 5:8
      └─attr("srcfile"): <srcfilecopy> @001
        ├─Enc: "unknown"
        ├─filename: "<scrubbed>"
        ├─fixedNewlines: TRUE
        ├─isFile: FALSE
        ├─lines<chr [5]>: "x <- functio...", "  # A long c...", "  # multiple...", ...
        ├─timestamp: "<scrubbed>"
        └─wd: "<scrubbed>"

# src() shows expression with single element

    Code
      x <- parse(text = "x + 1", keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 1:5
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines: "x + 1"
      │     ├─parseData<int [48]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 2:0
        └─attr("srcfile"): @001

# src() shows expression with multiple elements

    Code
      x <- parse(text = c("x + 1", "y + 2", "z + 3"), keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ ├─[[1]]: <srcref>
      │ │ ├─location: 1:1 - 1:5
      │ │ └─attr("srcfile"): <srcfilecopy> @001
      │ │   ├─Enc: "unknown"
      │ │   ├─filename: "<scrubbed>"
      │ │   ├─fixedNewlines: TRUE
      │ │   ├─isFile: FALSE
      │ │   ├─lines<chr [3]>: "x + 1", "y + 2", "z + 3"
      │ │   ├─parseData<int [144]>: 1, 1, 1, ...
      │ │   ├─timestamp: "<scrubbed>"
      │ │   └─wd: "<scrubbed>"
      │ ├─[[2]]: <srcref>
      │ │ ├─location: 2:1 - 2:5
      │ │ └─attr("srcfile"): @001
      │ └─[[3]]: <srcref>
      │   ├─location: 3:1 - 3:5
      │   └─attr("srcfile"): @001
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 4:0
        └─attr("srcfile"): @001

# src() shows expression with nested block and wholeSrcref

    Code
      x <- parse(text = "{\n  1\n}", keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 3:1
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines<chr [3]>: "{", "  1", "}"
      │     ├─parseData<int [40]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      ├─attr("wholeSrcref"): <srcref>
      │ ├─location: 1:0 - 4:0
      │ └─attr("srcfile"): @001
      └─[[1]]: <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:1 - 1:1
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 2:3 - 2:3
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 3:1
          └─attr("srcfile"): @001

# src() shows nested block element directly

    Code
      x <- parse(text = "{\n  1\n}", keep.source = TRUE)
      src(x[[1]])
    Output
      <{>
      ├─attr("srcref"): <list>
      │ ├─[[1]]: <srcref>
      │ │ ├─location: 1:1 - 1:1
      │ │ └─attr("srcfile"): <srcfilecopy> @001
      │ │   ├─Enc: "unknown"
      │ │   ├─filename: "<scrubbed>"
      │ │   ├─fixedNewlines: TRUE
      │ │   ├─isFile: FALSE
      │ │   ├─lines<chr [3]>: "{", "  1", "}"
      │ │   ├─parseData<int [40]>: 1, 1, 1, ...
      │ │   ├─timestamp: "<scrubbed>"
      │ │   └─wd: "<scrubbed>"
      │ └─[[2]]: <srcref>
      │   ├─location: 2:3 - 2:3
      │   └─attr("srcfile"): @001
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 3:1
        └─attr("srcfile"): @001

# src() shows block with srcref list and wholeSrcref

    Code
      x <- parse(text = "{\n  a <- 1\n  b <- 2\n}", keep.source = TRUE)
      src(x[[1]])
    Output
      <{>
      ├─attr("srcref"): <list>
      │ ├─[[1]]: <srcref>
      │ │ ├─location: 1:1 - 1:1
      │ │ └─attr("srcfile"): <srcfilecopy> @001
      │ │   ├─Enc: "unknown"
      │ │   ├─filename: "<scrubbed>"
      │ │   ├─fixedNewlines: TRUE
      │ │   ├─isFile: FALSE
      │ │   ├─lines<chr [4]>: "{", "  a <- 1", "  b <- 2", ...
      │ │   ├─parseData<int [120]>: 1, 1, 1, ...
      │ │   ├─timestamp: "<scrubbed>"
      │ │   └─wd: "<scrubbed>"
      │ ├─[[2]]: <srcref>
      │ │ ├─location: 2:3 - 2:8
      │ │ └─attr("srcfile"): @001
      │ └─[[3]]: <srcref>
      │   ├─location: 3:3 - 3:8
      │   └─attr("srcfile"): @001
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 4:1
        └─attr("srcfile"): @001

# src() shows single srcref

    Code
      x <- parse(text = "x + 1", keep.source = TRUE)
      sr <- attr(x, "srcref")[[1]]
      src(sr)
    Output
      <srcref>
      ├─location: 1:1 - 1:5
      └─attr("srcfile"): <srcfilecopy> @001
        ├─Enc: "unknown"
        ├─filename: "<scrubbed>"
        ├─fixedNewlines: TRUE
        ├─isFile: FALSE
        ├─lines: "x + 1"
        ├─parseData<int [48]>: 1, 1, 1, ...
        ├─timestamp: "<scrubbed>"
        └─wd: "<scrubbed>"

# src() shows list of srcrefs with count

    Code
      x <- parse(text = c("x + 1", "y + 2"), keep.source = TRUE)
      sr_list <- attr(x, "srcref")
      src(sr_list)
    Output
      <list>
      ├─[[1]]: <srcref>
      │ ├─location: 1:1 - 1:5
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: FALSE
      │   ├─lines<chr [2]>: "x + 1", "y + 2"
      │   ├─parseData<int [96]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─[[2]]: <srcref>
        ├─location: 2:1 - 2:5
        └─attr("srcfile"): @001

# src() reveals srcref list structure with index notation

    Code
      with_srcref("x <- quote(function() { 1 })")
      src(x[[3]])
    Output
      <{>
      ├─attr("srcref"): <list>
      │ ├─[[1]]: <srcref>
      │ │ ├─location: 1:23 - 1:23
      │ │ └─attr("srcfile"): <srcfilecopy> @001
      │ │   ├─Enc: "unknown"
      │ │   ├─filename: "<scrubbed>"
      │ │   ├─fixedNewlines: TRUE
      │ │   ├─isFile: TRUE
      │ │   ├─lines: "x <- quote(function() { 1 })"
      │ │   ├─parseData<int [144]>: 1, 1, 1, ...
      │ │   ├─timestamp: "<scrubbed>"
      │ │   └─wd: "<scrubbed>"
      │ └─[[2]]: <srcref>
      │   ├─location: 1:25 - 1:25
      │   └─attr("srcfile"): @001
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 1:27
        └─attr("srcfile"): @001

# src() handles srcrefs nested in language calls

    Code
      x <- parse(text = "foo({ if (1) bar({ 2 }) })", keep.source = TRUE)
      src(x, max_depth = 10)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 1:26
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines: "foo({ if (1) bar({ 2 }) })"
      │     ├─parseData<int [192]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      ├─attr("wholeSrcref"): <srcref>
      │ ├─location: 1:0 - 2:0
      │ └─attr("srcfile"): @001
      └─[[1]][[2]]: <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:5 - 1:5
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 1:7 - 1:23
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        ├─attr("wholeSrcref"): <srcref>
        │ ├─location: 1:0 - 1:25
        │ └─attr("srcfile"): @001
        └─[[2]][[3]][[2]]: <{>
          ├─attr("srcref"): <list>
          │ ├─[[1]]: <srcref>
          │ │ ├─location: 1:18 - 1:18
          │ │ └─attr("srcfile"): @001
          │ └─[[2]]: <srcref>
          │   ├─location: 1:20 - 1:20
          │   └─attr("srcfile"): @001
          ├─attr("srcfile"): @001
          └─attr("wholeSrcref"): <srcref>
            ├─location: 1:0 - 1:22
            └─attr("srcfile"): @001

# src() handles srcrefs nested in function bodies

    Code
      with_srcref("f <- function() foo({ if (1) bar({ 2 }) })")
      src(f, max_depth = 10)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:6 - 1:42
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines: "f <- function() foo({ if (1) bar..."
      │   ├─parseData<int [256]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <language>
        └─[[2]]: <{>
          ├─attr("srcref"): <list>
          │ ├─[[1]]: <srcref>
          │ │ ├─location: 1:21 - 1:21
          │ │ └─attr("srcfile"): @001
          │ └─[[2]]: <srcref>
          │   ├─location: 1:23 - 1:39
          │   └─attr("srcfile"): @001
          ├─attr("srcfile"): @001
          ├─attr("wholeSrcref"): <srcref>
          │ ├─location: 1:0 - 1:41
          │ └─attr("srcfile"): @001
          └─[[2]][[3]][[2]]: <{>
            ├─attr("srcref"): <list>
            │ ├─[[1]]: <srcref>
            │ │ ├─location: 1:34 - 1:34
            │ │ └─attr("srcfile"): @001
            │ └─[[2]]: <srcref>
            │   ├─location: 1:36 - 1:36
            │   └─attr("srcfile"): @001
            ├─attr("srcfile"): @001
            └─attr("wholeSrcref"): <srcref>
              ├─location: 1:0 - 1:38
              └─attr("srcfile"): @001

# src() currently shows duplicate srcfile objects

    Code
      f <- simple_function_with_srcref()
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:14 - 3:1
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines<chr [3]>: "test_func <-...", "  x + y", "}"
      │   ├─parseData<int [160]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:29 - 1:29
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 2:3 - 2:7
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 3:1
          └─attr("srcfile"): @001

# src() shows many duplicate srcfiles in nested expression

    Code
      x <- parse(text = "{\n  1\n  2\n}", keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 4:1
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines<chr [4]>: "{", "  1", "  2", ...
      │     ├─parseData<int [56]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      ├─attr("wholeSrcref"): <srcref>
      │ ├─location: 1:0 - 5:0
      │ └─attr("srcfile"): @001
      └─[[1]]: <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:1 - 1:1
        │ │ └─attr("srcfile"): @001
        │ ├─[[2]]: <srcref>
        │ │ ├─location: 2:3 - 2:3
        │ │ └─attr("srcfile"): @001
        │ └─[[3]]: <srcref>
        │   ├─location: 3:3 - 3:3
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 4:1
          └─attr("srcfile"): @001

# src() handles empty block

    Code
      x <- parse(text = "{}", keep.source = TRUE)
      src(x[[1]])
    Output
      <{>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 1:1
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines: "{}"
      │     ├─parseData<int [24]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      └─attr("wholeSrcref"): <srcref>
        ├─location: 1:0 - 1:2
        └─attr("srcfile"): @001

# src() handles function without arguments

    Code
      with_srcref("f <- function() { NULL }")
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:6 - 1:24
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines: "f <- function() { NULL }"
      │   ├─parseData<int [104]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:17 - 1:17
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 1:19 - 1:22
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 1:24
          └─attr("srcfile"): @001

# src() handles if statement with blocks

    Code
      x <- parse(text = "if (TRUE) { 1 } else { 2 }", keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 1:26
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines: "if (TRUE) { 1 } else { 2 }"
      │     ├─parseData<int [136]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      ├─attr("wholeSrcref"): <srcref>
      │ ├─location: 1:0 - 2:0
      │ └─attr("srcfile"): @001
      ├─[[1]][[3]]: <{>
      │ ├─attr("srcref"): <list>
      │ │ ├─[[1]]: <srcref>
      │ │ │ ├─location: 1:11 - 1:11
      │ │ │ └─attr("srcfile"): @001
      │ │ └─[[2]]: <srcref>
      │ │   ├─location: 1:13 - 1:13
      │ │   └─attr("srcfile"): @001
      │ ├─attr("srcfile"): @001
      │ └─attr("wholeSrcref"): <srcref>
      │   ├─location: 1:0 - 1:15
      │   └─attr("srcfile"): @001
      └─[[1]][[4]]: <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:22 - 1:22
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 1:24 - 1:24
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 1:26
          └─attr("srcfile"): @001

# src() respects show_source_lines parameter

    Code
      f <- simple_function_with_srcref()
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:14 - 3:1
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines<chr [3]>: "test_func <-...", "  x + y", "}"
      │   ├─parseData<int [160]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:29 - 1:29
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 2:3 - 2:7
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        └─attr("wholeSrcref"): <srcref>
          ├─location: 1:0 - 3:1
          └─attr("srcfile"): @001

# src() shows expression with multiple nested blocks

    Code
      x <- parse(text = "{\n  {\n    1\n  }\n  {\n    2\n  }\n}", keep.source = TRUE)
      src(x)
    Output
      <expression>
      ├─attr("srcref"): <list>
      │ └─[[1]]: <srcref>
      │   ├─location: 1:1 - 8:1
      │   └─attr("srcfile"): <srcfilecopy> @001
      │     ├─Enc: "unknown"
      │     ├─filename: "<scrubbed>"
      │     ├─fixedNewlines: TRUE
      │     ├─isFile: FALSE
      │     ├─lines<chr [8]>: "{", "  {", "    1", ...
      │     ├─parseData<int [104]>: 1, 1, 1, ...
      │     ├─timestamp: "<scrubbed>"
      │     └─wd: "<scrubbed>"
      ├─attr("srcfile"): @001
      ├─attr("wholeSrcref"): <srcref>
      │ ├─location: 1:0 - 9:0
      │ └─attr("srcfile"): @001
      └─[[1]]: <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:1 - 1:1
        │ │ └─attr("srcfile"): @001
        │ ├─[[2]]: <srcref>
        │ │ ├─location: 2:3 - 4:3
        │ │ └─attr("srcfile"): @001
        │ └─[[3]]: <srcref>
        │   ├─location: 5:3 - 7:3
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        ├─attr("wholeSrcref"): <srcref>
        │ ├─location: 1:0 - 8:1
        │ └─attr("srcfile"): @001
        ├─[[2]]: <{>
        │ ├─attr("srcref"): <list>
        │ │ ├─[[1]]: <srcref>
        │ │ │ ├─location: 2:3 - 2:3
        │ │ │ └─attr("srcfile"): @001
        │ │ └─[[2]]: <srcref>
        │ │   ├─location: 3:5 - 3:5
        │ │   └─attr("srcfile"): @001
        │ ├─attr("srcfile"): @001
        │ └─attr("wholeSrcref"): <srcref>
        │   ├─location: 1:0 - 4:3
        │   └─attr("srcfile"): @001
        └─[[3]]: <{>
          ├─attr("srcref"): <list>
          │ ├─[[1]]: <srcref>
          │ │ ├─location: 5:3 - 5:3
          │ │ └─attr("srcfile"): @001
          │ └─[[2]]: <srcref>
          │   ├─location: 6:5 - 6:5
          │   └─attr("srcfile"): @001
          ├─attr("srcfile"): @001
          └─attr("wholeSrcref"): <srcref>
            ├─location: 1:0 - 7:3
            └─attr("srcfile"): @001

# src() shows function with nested block in body

    Code
      with_srcref("f <- function(x) {\n  if (x) {\n    1\n  }\n}")
      src(f)
    Output
      <closure>
      ├─attr("srcref"): <srcref>
      │ ├─location: 1:6 - 5:1
      │ └─attr("srcfile"): <srcfilecopy> @001
      │   ├─Enc: "unknown"
      │   ├─filename: "<scrubbed>"
      │   ├─fixedNewlines: TRUE
      │   ├─isFile: TRUE
      │   ├─lines<chr [5]>: "f <- functio...", "  if (x) {", "    1", ...
      │   ├─parseData<int [184]>: 1, 1, 1, ...
      │   ├─timestamp: "<scrubbed>"
      │   └─wd: "<scrubbed>"
      └─body(): <{>
        ├─attr("srcref"): <list>
        │ ├─[[1]]: <srcref>
        │ │ ├─location: 1:18 - 1:18
        │ │ └─attr("srcfile"): @001
        │ └─[[2]]: <srcref>
        │   ├─location: 2:3 - 4:3
        │   └─attr("srcfile"): @001
        ├─attr("srcfile"): @001
        ├─attr("wholeSrcref"): <srcref>
        │ ├─location: 1:0 - 5:1
        │ └─attr("srcfile"): @001
        └─[[2]][[3]]: <{>
          ├─attr("srcref"): <list>
          │ ├─[[1]]: <srcref>
          │ │ ├─location: 2:10 - 2:10
          │ │ └─attr("srcfile"): @001
          │ └─[[2]]: <srcref>
          │   ├─location: 3:5 - 3:5
          │   └─attr("srcfile"): @001
          ├─attr("srcfile"): @001
          └─attr("wholeSrcref"): <srcref>
            ├─location: 1:0 - 4:3
            └─attr("srcfile"): @001

# can print complex expression

    Code
      ast(!!x)
    Output
      █─`function` 
      ├─█─x = `` 
      ├─█─`if` 
      │ ├─█─`>` 
      │ │ ├─x 
      │ │ └─1 
      │ └─█─f 
      │   ├─█─`$` 
      │   │ ├─y 
      │   │ └─x 
      │   ├─"x" 
      │   └─█─g 
      └─<inline srcref> 

# can print complex expression without unicode

    Code
      ast(!!x)
    Output
      o-`function` 
      +-o-x = `` 
      +-o-`if` 
      | +-o-`>` 
      | | +-x 
      | | \-1 
      | \-o-f 
      |   +-o-`$` 
      |   | +-y 
      |   | \-x 
      |   +-"x" 
      |   \-o-g 
      \-<inline srcref> 

# can print scalar expressions nicely

    Code
      ast(!!x)
    Output
      o-list 
      +-logical = o-c 
      |           +-FALSE 
      |           +-TRUE 
      |           \-NA 
      +-integer = 1L 
      +-double = 1 
      +-character = "a" 
      \-complex = 1i 

# snapshots environment binding types

    Code
      print(sxp(e))
    Output
      [1] <ENVSXP> ()
        active <CLOSXP: active>
        forced <PROMSXP>
        delayed <PROMSXP>
        missing <missing>
        value [2] <REALSXP[1]> ()
        _enclos [3] <ENVSXP: empty> ()
    Code
      print(sxp(e, expand = "environment", max_depth = 6L))
    Output
      [1] <ENVSXP> ()
        active <CLOSXP: active>
        _fn [2] <CLOSXP> ()
          _formals <NILSXP>
          _body [4] <REALSXP[1]> ()
          _env [5] <ENVSXP> ()
            e [1]
            _enclos [6] <ENVSXP> ()
              _enclos [7] <ENVSXP> ()
                _enclos [8] <ENVSXP> ()
                  ...
          _attrib [9] <LISTSXP> ()
            srcref [10] <INTSXP[8]> (object )
              _attrib [11] <LISTSXP> ()
                srcfile [12] <ENVSXP> (object )
                  ...
                class [13] <STRSXP[1]> ()
                  ...
        forced <PROMSXP>
        _value [14] <REALSXP[1]> ()
        _code [15] <LANGSXP> ()
          ...
        delayed <PROMSXP>
        _code [16] <LANGSXP> ()
          ...
        _env [5]
        missing <missing>
        value [17] <REALSXP[1]> ()
        _enclos [18] <ENVSXP: empty> ()

# can inspect all atomic vectors

    Code
      sxp(x)
    Output
      [1] <VECSXP[6]> ()
        [2] <LGLSXP[1]> ()
        [3] <INTSXP[1]> ()
        [4] <REALSXP[1]> ()
        [5] <STRSXP[1]> ()
        [6] <CPLXSXP[1]> ()
        [7] <RAWSXP[1]> ()

# can inspect functions

    Code
      sxp(f)
    Output
      [1] <CLOSXP> ()
        _formals [2] <LISTSXP> ()
          x [3] <SYMSXP: > ()
          y [4] <REALSXP[1]> ()
          ... [3]
        _body [5] <LANGSXP> ()
          ...
        _env [6] <ENVSXP: global> ()

# can inspect environments

    Code
      print(sxp(e2))
    Output
      [1] <ENVSXP> ()
        _enclos [2] <ENVSXP> ()
          x [3] <REALSXP[1]> ()
          y [2]
          _enclos [4] <ENVSXP: empty> ()
    Code
      print(sxp(e2, expand = "environment", max_depth = 5L))
    Output
      [1] <ENVSXP> ()
        _enclos [2] <ENVSXP> ()
          x [3] <REALSXP[1]> ()
          y [2]
          _enclos [4] <ENVSXP: empty> ()

# can expand altrep

    Code
      x <- 1:10
      print(sxp(x, expand = "altrep", max_depth = 4L))
    Output
      [1] <INTSXP[10]> (altrep )
        _class [2] <RAWSXP[144]> ()
          _attrib [3] <LISTSXP> ()
            [4] <SYMSXP: compact_intseq> ()
            [5] <SYMSXP: base> ()
            [6] <INTSXP[1]> ()
        _data1 [7] <REALSXP[3]> ()
        _data2 <NILSXP>

# can inspect cons cells

    Code
      cell <- new_node(1, 2)
      sxp(cell)
    Output
      [1] <LISTSXP> ()
        [2] <REALSXP[1]> ()
        _cdr [3] <REALSXP[1]> ()
    Code
      non_nil_terminated_list <- new_node(1, new_node(2, 3))
      sxp(non_nil_terminated_list)
    Output
      [1] <LISTSXP> ()
        [2] <REALSXP[1]> ()
        [3] <REALSXP[1]> ()
        _cdr [4] <REALSXP[1]> ()

# fix error message when `expand` argument contains invalid classes

    Code
      sxp(1, expand = "invalid_class")
    Condition
      Error in `sxp()`:
      ! `expand` must contain only values from: 'character', 'altrep', 'environment', 'call', 'bytecode'.

# basic list display

    Code
      x <- 1:10
      y <- list(x, x)
      ref(x, list(), list(x, x, x), list(a = x, b = x), letters)
    Output
      [1:0x001] <int> 
       
      █ [2:0x002] <list> 
       
      █ [3:0x003] <list> 
      ├─[1:0x001] 
      ├─[1:0x001] 
      └─[1:0x001] 
       
      █ [4:0x004] <named list> 
      ├─a = [1:0x001] 
      └─b = [1:0x001] 
       
      [5:0x005] <chr> 

# basic environment display

    Code
      e <- env(a = 1:10)
      e$b <- e$a
      e$c <- e
      ref(e)
    Output
      █ [1:0x001] <env> 
      ├─a = [2:0x002] <int> 
      ├─b = [2:0x002] 
      └─c = [1:0x001] 

# environment shows objects beginning with .

    Code
      e <- env(. = 1:10)
      ref(e)
    Output
      █ [1:0x001] <env> 
      └─. = [2:0x002] <int> 

# can display ref to global string pool on request

    Code
      ref(c("string", "string", "new string"), character = TRUE)
    Output
      █ [1:0x001] <chr> 
      ├─[2:0x002] <string: "string"> 
      ├─[2:0x002] 
      └─[3:0x003] <string: "new string"> 

# Array-like indices can be shown or hidden

    Code
      tree(list(a = "a", "b", "c"), index_unnamed = TRUE)
    Output
      <list>
      ├─a: "a"
      ├─2: "b"
      └─3: "c"

---

    Code
      tree(list(a = "a", "b", "c"), index_unnamed = FALSE)
    Output
      <list>
      ├─a: "a"
      ├─"b"
      └─"c"

# Atomic arrays have sensible defaults w/ truncation for longer than 10-elements

    Code
      tree(list(name = "vectored list", num_vec = 1:10, char_vec = letters))
    Output
      <list>
      ├─name: "vectored list"
      ├─num_vec<int [10]>: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
      └─char_vec<chr [26]>: "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", ...

---

    Code
      tree(list(name = "vectored list", num_vec = 1:10, char_vec = letters),
      hide_scalar_types = FALSE)
    Output
      <list>
      ├─name<chr [1]>: "vectored list"
      ├─num_vec<int [10]>: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
      └─char_vec<chr [26]>: "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", ...

# Large and multiline strings are handled gracefully

    Code
      long_strings <- list(`normal string` = "first element", `really long string` = paste(
        rep(letters, 4), collapse = ""), `vec of long strings` = c(
        "a long\nand multi\nline string element", "a fine length",
        "another long\nand also multi\nline string element"))
      tree(long_strings)
    Output
      <list>
      ├─normal string: "first element"
      ├─really long string: "abcdefghijklmnopqrstuvwxyzabcdef..."
      └─vec of long strings<chr [3]>: "a long↵and m...", "a fine length", "another long..."
    Code
      tree(long_strings, remove_newlines = FALSE)
    Output
      <list>
      ├─normal string: "first element"
      ├─really long string: "abcdefghijklmnopqrstuvwxyzabcdef..."
      └─vec of long strings<chr [3]>: "a long
      and m...", "a fine length", "another long..."

# Max depth and length can be enforced

    Code
      deep_list <- list(list(id = "b", val = 1, children = list(list(id = "b1", val = 2.5),
      list(id = "b2", val = 8, children = list(list(id = "b21", val = 4))))), list(
        id = "a", val = 2))
      tree(deep_list, max_depth = 1)
    Output
      <list>
      +-<list>...
      \-<list>...
    Code
      tree(deep_list, max_depth = 2)
    Output
      <list>
      +-<list>
      | +-id: "b"
      | +-val: 1
      | \-children: <list>...
      \-<list>
        +-id: "a"
        \-val: 2
    Code
      tree(deep_list, max_depth = 3)
    Output
      <list>
      +-<list>
      | +-id: "b"
      | +-val: 1
      | \-children: <list>
      |   +-<list>...
      |   \-<list>...
      \-<list>
        +-id: "a"
        \-val: 2
    Code
      tree(deep_list, max_length = 0)
    Output
      ... 
    Code
      tree(deep_list, max_length = 2)
    Output
      <list>
      +-<list>
      ... 
    Code
      tree(deep_list, max_depth = 1, max_length = 4)
    Output
      <list>
      +-<list>...
      \-<list>...

# Missing values are caught and printed properly

    Code
      tree(list(`null-element` = NULL, `NA-element` = NA))
    Output
      <list>
      ├─null-element: <NULL>
      └─NA-element: NA

# non-named elements in named list

    Code
      tree(list(a = 1, "el w/o id"))
    Output
      <list>
      ├─a: 1
      └─"el w/o id"

# Attributes are properly displayed as special children nodes

    Code
      list_w_attrs <- structure(list(structure(list(id = "a", val = 2), level = 2,
      name = "first child"), structure(list(id = "b", val = 1, children = list(list(
        id = "b1", val = 2.5))), level = 2, name = "second child", class = "custom-class"),
      level = "1", name = "root"))
      tree(list_w_attrs, show_attributes = TRUE)
    Output
      <list>
      ├─<list>
      │ ├─id: "a"
      │ ├─val: 2
      │ ├┄attr(,"names")<chr [2]>: "id", "val"
      │ ├┄attr(,"level"): 2
      │ └┄attr(,"name"): "first child"
      ├─S3<custom-class>
      │ ├─id: "b"
      │ ├─val: 1
      │ ├─children: <list>
      │ ┊ └─<list>
      │ ┊   ├─id: "b1"
      │ ┊   ├─val: 2.5
      │ ┊   └┄attr(,"names")<chr [2]>: "id", "val"
      │ ├┄attr(,"names")<chr [3]>: "id", "val", "children"
      │ ├┄attr(,"level"): 2
      │ ├┄attr(,"name"): "second child"
      │ └┄attr(,"class"): "custom-class"
      ├─level: "1"
      ├─name: "root"
      └┄attr(,"names")<chr [4]>: "", "", "level", "name"
    Code
      tree(list_w_attrs, show_attributes = FALSE)
    Output
      <list>
      ├─<list>
      │ ├─id: "a"
      │ └─val: 2
      ├─S3<custom-class>
      │ ├─id: "b"
      │ ├─val: 1
      │ └─children: <list>
      │   └─<list>
      │     ├─id: "b1"
      │     └─val: 2.5
      ├─level: "1"
      └─name: "root"

# Function arguments get printed

    Code
      tree(list(no_args = function() { }, few_args = function(a, b, c) { },
      lots_of_args = function(d, e, f, g, h, i, j, k, l, m, n, o, p) { }))
    Output
      <list>
      ├─no_args: function()
      ├─few_args: function(a, b, c)
      └─lots_of_args: function(d, e, f, g, h, ...)

# Handles expressions

    Code
      tree(list(a = quote(a), b = quote(a + 1), c = y ~ mx + b))
    Output
      <list>
      ├─a: <symbol> a
      ├─b: <language> a + 1
      └─c: S3<formula> y ~ mx + b

# Hidden lists dont cause infinite recursion

    Code
      tree(package_version("1.2.3"))
    Output
      S3<package_version/numeric_version>
      └─<int [3]>1, 2, 3

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/address.R
\name{obj_addr}
\alias{obj_addr}
\alias{obj_addrs}
\title{Find memory location of objects and their children.}
\usage{
obj_addr(x)

obj_addrs(x)
}
\arguments{
\item{x}{An object}
}
\description{
\code{obj_addr()} gives the address of the value that \code{x} points to;
\code{obj_addrs()} gives the address of the components the list,
environment, and character vector \code{x} point to.
}
\details{
\code{obj_addr()} has been written in such away that it avoids taking
references to an object.
}
\examples{
# R creates copies lazily
x <- 1:10
y <- x
obj_addr(x) == obj_addr(y)

y[1] <- 2L
obj_addr(x) == obj_addr(y)

y <- runif(10)
obj_addr(y)
z <- list(y, y)
obj_addrs(z)

y[2] <- 1.0
obj_addrs(z)
obj_addr(y)

# The address of an object is different every time you create it:
obj_addr(1:10)
obj_addr(1:10)
obj_addr(1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ref.R
\name{ref}
\alias{ref}
\title{Display tree of references}
\usage{
ref(..., character = FALSE)
}
\arguments{
\item{...}{One or more objects}

\item{character}{If \code{TRUE}, show references from character vector in to
global string pool}
}
\description{
This tree display focusses on the distinction between names and values.
For each reference-type object (lists, environments, and optional character
vectors), it displays the location of each component. The display
shows the connection between shared references using a locally unique id.
}
\examples{
x <- 1:100
ref(x)

y <- list(x, x, x)
ref(y)
ref(x, y)

e <- new.env()
e$e <- e
e$x <- x
e$y <- list(x, e)
ref(e)

# Can also show references to global string pool if requested
ref(c("x", "x", "y"))
ref(c("x", "x", "y"), character = TRUE)
}
\seealso{
Other object inspectors: 
\code{\link{ast}()},
\code{\link{src}()},
\code{\link{sxp}()}
}
\concept{object inspectors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sxp.R
\name{sxp}
\alias{sxp}
\title{Inspect an object}
\usage{
sxp(x, expand = character(), max_depth = 5L)
}
\arguments{
\item{x}{Object to inspect}

\item{expand}{Optionally, expand components of the true that are usually
suppressed. Use:
\itemize{
\item "character" to show underlying entries in the global string pool.
\item "environment" to show binding components without any side effects (e.g. promises or active bindings).
\item "altrep" to show the underlying data.
\item "call" to show the full AST (but \code{\link[=ast]{ast()}} is usually superior)
\item "bytecode" to show generated bytecode.
}}

\item{max_depth}{Maximum depth to recurse. Use \code{max_depth = Inf} (with care!)
to recurse as deeply as possible. Skipped elements will be shown as \code{...}.`}
}
\description{
\code{sxp(x)} is similar to \code{.Internal(inspect(x))}, recursing into the C data
structures underlying any R object. The main difference is the output is a
little more compact, it recurses fully, and avoids getting stuck in infinite
loops by using a depth-first search. It also returns a list that you can
compute with, and carefully uses colour to highlight the most important
details.
}
\details{
The name \code{sxp} comes from \code{SEXP}, the name of the C data structure that
underlies all R objects.
}
\examples{
x <- list(
  TRUE,
  1L,
  runif(100),
  "3"
)
sxp(x)

# Expand "character" to see underlying CHARSXP entries in the global
# string pool
x <- c("banana", "banana", "apple", "banana")
sxp(x)
sxp(x, expand = "character")

# Expand altrep to see underlying data
x <- 1:10
sxp(x)
sxp(x, expand = "altrep")

# Expand environments to see promise expressions without forcing
e <- new.env(parent = emptyenv())
delayedAssign("x", 1 + 1, assign.env = e)

sxp(e)
sxp(e, expand = "environment")
}
\seealso{
Other object inspectors: 
\code{\link{ast}()},
\code{\link{ref}()},
\code{\link{src}()}
}
\concept{object inspectors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/src.R
\name{src}
\alias{src}
\title{Display tree of source references}
\usage{
src(x, max_depth = 5L, max_length = 100L, ...)
}
\arguments{
\item{x}{An R object with source references. Can be:
\itemize{
\item A \code{srcref} object
\item A list of \code{srcref} objects
\item A expression vector with attached source references
\item An evaluated closure with attached source references
\item A quoted call with attached source references
}}

\item{max_depth}{Maximum depth to traverse nested structures (default 5)}

\item{max_length}{Maximum number of srcref nodes to display (default 100)}

\item{...}{Additional arguments passed to \code{\link[=tree]{tree()}}}
}
\value{
Returns a structured list containing the source reference
information. Print it to view the formatted tree.
}
\description{
View source reference metadata attached to R objects in a tree structure.
Shows source file information, line/column locations, and lines of source code.
}
\section{Overview}{


Source references are made of two kinds of objects:
\itemize{
\item \code{srcref} objects, which contain information about a specific
location within the source file, such as the line and column numbers.
\item \code{srcfile} objects, which contain metadata about the source file
such as its name, path, and encoding.
}
\subsection{Where and when are source references created?}{

Ultimately the R parser creates source references. The main two entry points
to the parser are:
\itemize{
\item The R function \code{parse()}.
\item The frontend hook \code{ReadConsole}, which powers the console input parser in
the R CLI and in IDEs. This C-level parser can also be accessed from C code
via \code{R_ParseVector()}.
}

In principle, anything that calls \code{parse()} may create source references, but
here are the important direct and indirect callers:
\itemize{
\item \code{source()} and \code{sys.source()} which parse and evaluate code.
\item \code{loadNamespace()} calls \code{sys.source()} when loading a \emph{source} package:
\url{https://github.com/r-devel/r-svn/blob/acd196be/src/library/base/R/namespace.R#L573}.
\item \verb{R CMD install} creates a lazy-load database from a source package.
The first step is to call \code{loadNamespace()}:
\url{https://github.com/r-devel/r-svn/blob/acd196be/src/library/tools/R/makeLazyLoad.R#L32}
}

By default source references are not created but can be enabled by:
\itemize{
\item Passing \code{keep.source = TRUE} explicitly to \code{parse()}, \code{source()},
\code{sys.source()}, or \code{loadNamespace()}.
\item Setting \code{options(keep.source = TRUE)}. This affects the default arguments
of the aforementioned functions, as well as the console input parser.
In interactive sessions, \code{keep.source} is set to \code{TRUE} by default:
\url{https://github.com/r-devel/r-svn/blob/3a4745af/src/library/profile/Common.R#L26}.
\item Setting \code{options(keep.source.pkgs = TRUE)}. This affects loading a package
from source, and installing a package from source.
}
}

\subsection{\code{srcref} objects}{

\code{srcref} objects are compact integer vectors describing a character range
in a source. It records start/end lines and byte/column positions and,
optionally, the parsed-line numbers if \verb{#line} directives were used.

Lengths of 4, 6, or 8 are allowed:
\itemize{
\item 4: basic (first_line, first_byte, last_line, last_byte). Byte positions
are within the line.
\item 6: adds columns in Unicode codepoints (first_col, last_col)
\item 8: adds parsed-line numbers (first_parsed, last_parsed)
}

The "column" information does not represent grapheme clusters, but Unicode
codepoints. The column cursor is incremented at every UTF-8 lead byte and
there is no support for encodings other than UTF-8.

The srcref columns are right-boundary positions, meaning that for an
expression starting at the start of a line, the column will be 1.
\code{wholeSrcref} (see below) on the other hand starts at 0, before the first
character. It might also end 1 character after the last srcref column.

They are attached as attributes (e.g. \code{attr(x, "srcref")} or \code{attr(x, "wholeSrcref")}), possibly wrapped in a list, to the following objects:
\itemize{
\item Expression vectors returned by \code{parse()} (wrapped in a list)
\item Quoted function calls (unwrapped)
\item Quoted \verb{\{} calls (wrapped in a list). This is crucial for debugging: when R
steps through brace lists, the srcref for the current expression is saved to
a global variable (\code{R_Srcref}) so the IDE knows exactly where execution is
paused. See: \url{https://github.com/r-devel/r-svn/blob/fa0b47c5/src/main/eval.c#L2986}.
\item Evaluated closures (unwrapped)
}

They have a \code{srcfile} attribute that points to the source file.

Methods:
\itemize{
\item \code{as.character()}: Retrieves relevant source lines from the \code{srcfile}
reference.
}
\subsection{\code{wholeSrcref} attributes}{

These are \code{srcref} objects stored in the \code{wholeSrcref} attributes of:
\itemize{
\item Expression vectors returned by \code{parse()}, which seems to be the intended
usage.
\item \verb{\{} calls, which seems unintended.
}

For expression vectors, the \code{wholeSrcref} spans from the first position
to the last position and represents the entire document. For braces, they
span from the first position to the location of the closing brace. There is
no way to know the location of the opening brace without reparsing, which
seems odd. It's probably an overlook from \code{xxexprlist()} calling
\code{attachSrcrefs()} in
\url{https://github.com/r-devel/r-svn/blob/52affc16/src/main/gram.y#L1380}. That
function is also called at the end of parsing, where it's intended for the
\code{wholeSrcref} attribute to be attached.

For evaluated closures, the \code{wholeSrcref} attribute on the body has the same
unreliable start positions as \verb{\{} nodes.
}

}

\subsection{\code{srcfile} objects}{

\code{srcfile} objects are environments representing information about a
source file that a source reference points to. They typically refer to
a file on disk and store the filename, working directory, a timestamp,
and encoding information.

While it is possible to create bare \code{srcfile} objects, specialized subclasses
are much more common.
\subsection{\code{srcfile}}{

A bare \code{srcfile} object does not contain any data apart from the file path.
It lazily loads lines from the file on disk, without any caching.

Fields common to all \code{srcfile} objects:
\itemize{
\item \code{filename}: The filename of the source file. If relative, the path is
resolved against \code{wd}.
\item \code{wd}: The working directory (\code{getwd()}) at the time the srcfile was created,
generally at the time of parsing).
\item \code{timestamp}: The timestamp of the source file. Retrieved from \code{filename}
with \code{file.mtime()}.
\item \code{encoding}: The encoding of the source file.
\item \code{Enc}: The encoding of output lines. Used by \code{getSrcLines()}, which
calls \code{iconv()} when \code{Enc} does not match \code{encoding}.
\item \code{parseData} (optional): Parser information saved when \code{keep.source.data} is
set to \code{TRUE}.
}

Implementations:
\itemize{
\item \code{print()} and \code{summary()} to print information about the source file.
\item \code{open()} and \code{close()} to access the underlying file as a connection.
}

Helpers:
\itemize{
\item \code{getSrcLines()}: Retrieves source lines from a \code{srcfile}.
}
}

\subsection{\code{srcfilecopy}}{

A \code{srcfilecopy} stores the actual source lines in memory in \verb{$lines}.
\code{srcfilecopy} is useful when the original file may change or does not
exist, because it preserves the exact text used by the parser.

This type of srcfile is the most common. It's created by:
\itemize{
\item The R-level \code{parse()} function when \code{text} is supplied:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Creates a `"<text>"` non-file `srcfilecopy`
parse(text = "...", keep.source = TRUE)
}\if{html}{\out{</div>}}
\item The console's input parser when \code{getOption("keep.source")} is \code{TRUE}.
\item \code{sys.source()} when \code{keep.source = TRUE}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{sys.source(file, keep.source = TRUE)
}\if{html}{\out{</div>}}

The \code{srcfilecopy} object is timestamped with the file's last modification time.
\url{https://github.com/r-devel/r-svn/blob/52affc16/src/library/base/R/source.R#L273-L276}
}

Fields:
\itemize{
\item \code{filename}: The filename of the source file. If \code{isFile} is \code{FALSE},
the field is non meaningful. For instance \code{parse(text = )} sets it to
\code{"<text>"}, and the console input parser sets it to \code{""}.
\item \code{isFile}: A logical indicating whether the source file exists.
\item \code{fixedNewlines}: If \code{TRUE}, \code{lines} is a character vector of lines with
no embedded \verb{\\n} characters. The \code{getSrcLines()} helper regularises \code{lines}
in this way and sets \code{fixedNewlines} to \code{TRUE}.
}

Note that the C-level parser (used directly mainly when parsing console input)
does not call the R-level constructor and only instantiates the \code{filename}
(set to \code{""}) and \code{lines} fields.
}

\subsection{\code{srcfilealias}}{

This object wraps an existing \code{srcfile} object (stored in \code{original}).  It
allows exposing a different \code{filename} while delegating the open/close/get
lines operations to the \code{srcfile} stored in \code{original}.

The typical way aliases are created is via \verb{#line *line* *filename*}
directives where the optional \verb{*filename*} argument is supplied. These
directives remap the srcref and srcfile of parsed code to a different
location, for example from a temporary file or generated file to the original
location on disk.

Created by \code{install.packages()} when installing a \emph{source} package with \code{keep.source.pkgs} set to \code{TRUE} (see
\url{https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/install.R#L545}), but
\href{https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/admin.R#L308}{only when}:
\itemize{
\item \code{Encoding} was supplied in \code{DESCRIPTION}
\item The system locale is not "C" or "POSIX".
}

The source files are converted to the encoding of the system locale, then
collated in a single source file with \verb{#line} directives mapping them to their
original file names (with full paths):
\url{https://github.com/r-devel/r-svn/blob/52affc16/src/library/tools/R/admin.R#L342}.

Note that the \code{filename} of the \code{original} srcfile incorrectly points to the
package path in the install destination.

Fields:
\itemize{
\item \code{filename}: The virtual file name (or full path) of the parsed code.
\item \code{original}: The actual \code{srcfile} the code was parsed from.
}
}

}
}

\seealso{
\itemize{
\item \code{\link[=srcfile]{srcfile()}}: Base documentation for \code{srcref} and \code{srcfile} objects.
\item \code{\link[=getParseData]{getParseData()}}: Parse information stored when \code{keep.source.data} is \code{TRUE}.
\item Source References (R Journal): \url{https://journal.r-project.org/articles/RJ-2010-010/}
}

Other object inspectors:
\code{\link[=ast]{ast()}},
\code{\link[=ref]{ref()}},
\code{\link[=sxp]{sxp()}}
}
\concept{object inspectors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ast.R
\name{ast}
\alias{ast}
\title{Display the abstract syntax tree}
\usage{
ast(x)
}
\arguments{
\item{x}{An expression to display. Input is automatically quoted,
use \verb{!!} to unquote if you have already captured an expression object.}
}
\description{
This is a useful alternative to \code{str()} for expression objects.
}
\examples{
# Leaves
ast(1)
ast(x)

# Simple calls
ast(f())
ast(f(x, 1, g(), h(i())))
ast(f()())
ast(f(x)(y))

ast((x + 1))

# Displaying expression already stored in object
x <- quote(a + b + c)
ast(x)
ast(!!x)

# All operations have this same structure
ast(if (TRUE) 3 else 4)
ast(y <- x * 10)
ast(function(x = 1, y = 2) { x + y } )

# Operator precedence
ast(1 * 2 + 3)
ast(!1 + !1)
}
\seealso{
Other object inspectors: 
\code{\link{ref}()},
\code{\link{src}()},
\code{\link{sxp}()}
}
\concept{object inspectors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mem.R
\name{mem_used}
\alias{mem_used}
\title{How much memory is currently used by R?}
\usage{
mem_used()
}
\description{
\code{mem_used()} wraps around \code{gc()} and returns the exact number of bytes
currently used by R. Note that changes will not match up exactly to
\code{\link[=obj_size]{obj_size()}} as session specific state (e.g. \link{.Last.value}) adds minor
variations.
}
\examples{
prev_m <- 0; m <- mem_used(); m - prev_m

x <- 1:1e6
prev_m <- m; m <- mem_used(); m - prev_m
obj_size(x)

rm(x)
prev_m <- m; m <- mem_used(); m - prev_m

prev_m <- m; m <- mem_used(); m - prev_m
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{tree_label}
\alias{tree_label}
\title{Build element or node label in tree}
\usage{
tree_label(x, opts)
}
\arguments{
\item{x}{A tree like object (list, etc.)}

\item{opts}{A list of options that directly mirrors the named arguments of
\link{tree}. E.g. \code{list(val_printer = crayon::red)} is equivalent to
\code{tree(..., val_printer = crayon::red)}.}
}
\description{
These methods control how the value of a given node is printed. New methods
can be added if support is needed for a novel class
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cst.R
\name{cst}
\alias{cst}
\title{Call stack tree}
\usage{
cst()
}
\description{
Shows the relationship between calls on the stack. This function
combines the results of \code{\link[=sys.calls]{sys.calls()}} and \code{\link[=sys.parents]{sys.parents()}} yielding a display
that shows how frames on the call stack are related.
}
\examples{
# If all evaluation is eager, you get a single tree
f <- function() g()
g <- function() h()
h <- function() cst()
f()

# You get multiple trees with delayed evaluation
try(f())

# Pay attention to the first element of each subtree: each
# evaluates the outermost call
f <- function(x) g(x)
g <- function(x) h(x)
h <- function(x) x
try(f(cst()))

# With a little ingenuity you can use it to see how NSE
# functions work in base R
with(mtcars, {cst(); invisible()})
invisible(subset(mtcars, {cst(); cyl == 0}))

# You can also get unusual trees by evaluating in frames
# higher up the call stack
f <- function() g()
g <- function() h()
h <- function() eval(quote(cst()), parent.frame(2))
f()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/size.R
\name{obj_size}
\alias{obj_size}
\alias{obj_sizes}
\title{Calculate the size of an object.}
\usage{
obj_size(..., env = parent.frame())

obj_sizes(..., env = parent.frame())
}
\arguments{
\item{...}{Set of objects to compute size.}

\item{env}{Environment in which to terminate search. This defaults to the
current environment so that you don't include the size of objects that
are already stored elsewhere.

Regardless of the value here, \code{obj_size()} never looks past the
global or base environments.}
}
\value{
An estimate of the size of the object, in bytes.
}
\description{
\code{obj_size()} computes the size of an object or set of objects;
\code{obj_sizes()} breaks down the individual contribution of multiple objects
to the total size.
}
\section{Compared to \code{object.size()}}{

Compared to \code{\link[=object.size]{object.size()}}, \code{obj_size()}:
\itemize{
\item Accounts for all types of shared values, not just strings in
the global string pool.
\item Includes the size of environments (up to \code{env})
\item Accurately measures the size of ALTREP objects.
}
}

\section{Environments}{

\code{obj_size()} attempts to take into account the size of the
environments associated with an object. This is particularly important
for closures and formulas, since otherwise you may not realise that you've
accidentally captured a large object. However, it's easy to over count:
you don't want to include the size of every object in every environment
leading back to the \code{\link[=emptyenv]{emptyenv()}}. \code{obj_size()} takes
a heuristic approach: it never counts the size of the global environment,
the base environment, the empty environment, or any namespace.

Additionally, the \code{env} argument allows you to specify another
environment at which to stop. This defaults to the environment from which
\code{obj_size()} is called to prevent double-counting of objects created
elsewhere.
}

\examples{
# obj_size correctly accounts for shared references
x <- runif(1e4)
obj_size(x)

z <- list(a = x, b = x, c = x)
obj_size(z)

# this means that object size is not transitive
obj_size(x)
obj_size(z)
obj_size(x, z)

# use obj_size() to see the unique contribution of each component
obj_sizes(x, z)
obj_sizes(z, x)
obj_sizes(!!!z)

# obj_size() also includes the size of environments
f <- function() {
  x <- 1:1e4
  a ~ b
}
obj_size(f())

#' # In R 3.5 and greater, `:` creates a special "ALTREP" object that only
# stores the first and last elements. This will make some vectors much
# smaller than you'd otherwise expect
obj_size(1:1e6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lobstr-package.R
\docType{package}
\name{lobstr-package}
\alias{lobstr}
\alias{lobstr-package}
\title{lobstr: Visualize R Data Structures with Trees}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A set of tools for inspecting and understanding R data structures inspired by str(). Includes ast() for visualizing abstract syntax trees, ref() for showing shared references, cst() for showing call stack trees, and obj_size() for computing object sizes.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://lobstr.r-lib.org/}
  \item \url{https://github.com/r-lib/lobstr}
  \item Report bugs at \url{https://github.com/r-lib/lobstr/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Authors:
\itemize{
  \item Hadley Wickham \email{hadley@posit.co}
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{tree}
\alias{tree}
\title{Pretty tree-like object printing}
\usage{
tree(
  x,
  ...,
  index_unnamed = FALSE,
  max_depth = 10L,
  max_length = 1000L,
  max_vec_len = 10L,
  show_environments = TRUE,
  hide_scalar_types = TRUE,
  val_printer = crayon::blue,
  class_printer = crayon::silver,
  show_attributes = FALSE,
  remove_newlines = TRUE,
  tree_chars = box_chars()
)
}
\arguments{
\item{x}{A tree like object (list, etc.)}

\item{...}{Ignored (used to force use of names)}

\item{index_unnamed}{Should children of containers without names have indices
used as stand-in?}

\item{max_depth}{How far down the tree structure should be printed. E.g. \code{1}
means only direct children of the root element will be shown. Useful for
very deep lists.}

\item{max_length}{How many elements should be printed? This is useful in case
you try and print an object with 100,000 items in it.}

\item{max_vec_len}{How many elements should be printed for vectors?}

\item{show_environments}{Should environments be treated like normal lists and
recursed into?}

\item{hide_scalar_types}{Should atomic scalars be printed with type and
length like vectors? E.g. \code{x <- "a"} would be shown as \verb{x<char [1]>: "a"}
instead of \code{x: "a"}.}

\item{val_printer}{Function that values get passed to before being drawn to
screen. Can be used to color or generally style output.}

\item{class_printer}{Same as \code{val_printer} but for the the class types of
non-atomic tree elements.}

\item{show_attributes}{Should attributes be printed as a child of the list or
avoided?}

\item{remove_newlines}{Should character strings with newlines in them have
the newlines removed? Not doing so will mess up the vertical flow of the
tree but may be desired for some use-cases if newline structure is
important to understanding object state.}

\item{tree_chars}{List of box characters used to construct tree. Needs
elements \verb{$h} for horizontal bar, \verb{$hd} for dotted horizontal bar, \verb{$v} for
vertical bar, \verb{$vd} for dotted vertical bar, \verb{$l} for l-bend, and \verb{$j} for
junction (or middle child).}
}
\value{
console output of structure
}
\description{
A cleaner and easier to read replacement for \code{str} for nested list-like
objects
}
\examples{

x <- list(
  list(id = "a", val = 2),
  list(
    id = "b",
    val = 1,
    children = list(
      list(id = "b1", val = 2.5),
      list(
        id = "b2",
        val = 8,
        children = list(
          list(id = "b21", val = 4)
        )
      )
    )
  ),
  list(
    id = "c",
    val = 8,
    children = list(
      list(id = "c1"),
      list(id = "c2", val = 1)
    )
  )
)

# Basic usage
tree(x)

# Even cleaner output can be achieved by not printing indices
tree(x, index_unnamed = FALSE)

# Limit depth if object is potentially very large
tree(x, max_depth = 2)

# You can customize how the values and classes are printed if desired
tree(x, val_printer = function(x) {
  paste0("_", x, "_")
})
}
